(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-Clang                                      *)
(*                                                                        *)
(*  Copyright (C) 2012-2017                                               *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(*                                                                        *)
(**************************************************************************)

open Intermediate_format
open Cabs
open Cil

let id x = x

let empty_qual n = { prequalification = []; decl_name = n }

let raw_block stmts = {blabels = []; battrs = []; bstmts = stmts}

let make_block env stmts =
  let loc = Convert_env.get_loc env in
  BLOCK(raw_block stmts,loc, loc)

let make_block_stmt env stmts =
  { stmt_ghost = false; stmt_node = make_block env stmts }

let is_static l = List.mem Static l

(* f(void) means no argument, not one argument of type void *)
let remove_void prm =
  match prm with
    [ { plain_type = Void } ] -> []
  | prm -> prm

let convert_cv = function
  | Const -> SpecCV CV_CONST
  | Volatile -> SpecCV CV_VOLATILE
  | Restrict -> SpecCV CV_RESTRICT
  | Static -> SpecStorage STATIC

let cv_to_attr = function
  | SpecCV CV_CONST -> "const",[]
  | SpecCV CV_VOLATILE -> "volatile",[]
  | SpecCV CV_RESTRICT -> "restrict", []
  | SpecStorage STATIC -> "static", []
  | _ -> "unknown_cv_specifier", []

(* creates an array[dim] of d. Because Cabs follows closely the C
   syntax, we have to take some precautions here. *)
let rec protect_array_type al dim d =
  match d with
  | JUSTBASE -> ARRAY(d,al,dim)
  | PARENTYPE (al1,d',al2) -> 
    PARENTYPE(al1,protect_array_type al dim d', al2)
  | ARRAY(d',al',dim') ->
    (* array dim of array dim' of d' is d' foo[dim][dim'] *)
    ARRAY(protect_array_type al dim d',al',dim')
  | PTR (al',d') ->
    (* array dim of ptr to d' is d' *foo[dim] *)
    PTR(al',protect_array_type al dim d')
    (* array dim of ptr to function returning d' is d' ( *foo[dim]()) *)
  | PROTO(d',args,variadic) ->
    PROTO(protect_array_type al dim d',args,variadic)

(* creates a *d. Similar issue as for protect_array_type. *)
let rec protect_ptr_type al d =
  match d with
  | JUSTBASE -> PTR(al,d)
  | PARENTYPE(al1,d',al2) -> PARENTYPE(al1,protect_ptr_type al d', al2)
  | ARRAY(d',al',dim') ->
    (* pointer to array dim of d is d ( *foo)[dim] *)
    ARRAY(PARENTYPE([],protect_ptr_type al d',[]),al',dim')
  | PTR(al',d') -> PTR(al', protect_ptr_type al d')
  | PROTO _ ->
    (* pointer to function is handled differently. Here, we have a function
       returning a pointer to something else. *)
    PTR (al, d)

let spec_type t = SpecType t

let spec_of_ikind s =
  Cil_types.(
    match s with
      | IBool -> [ SpecType Tbool ]
      | IChar -> [ SpecType Tchar ]
      | ISChar -> [ SpecType Tsigned; SpecType Tchar ]
      | IUChar -> [ SpecType Tunsigned; SpecType Tchar ]
      | IInt -> [ SpecType Tint ]
      | IUInt -> [ SpecType Tunsigned ]
      | IShort -> [ SpecType Tshort ]
      | IUShort -> [ SpecType Tunsigned; SpecType Tshort ]
      | ILong -> [ SpecType Tlong ]
      | IULong -> [ SpecType Tunsigned; SpecType Tlong ]
      | ILongLong -> [ SpecType Tlong; SpecType Tlong ]
      | IULongLong -> [ SpecType Tunsigned; SpecType Tlong; SpecType Tlong ]
  )

let make_integral_constant_kind k v =
  let v = Int64.to_string v in
  let s =
    match k with
      | IBool
      | IChar_s | ISChar | IWChar_s | IWSChar 
      | IChar | IWChar | IShort | IInt -> ""
      | IChar_u | IUChar | IChar16 | IChar32 | IWChar_u | IWUChar
      | IUShort | IUInt -> "U"
      | ILong -> "L"
      | IULong -> "UL"
      | ILongLong -> "LL"
      | IULongLong -> "ULL"
  in CONST_INT (v ^ s)

let local_reference env n =
  let var = VARIABLE n in
  match Convert_env.get_local_var env n with
    | LVReference _ | RVReference _ ->
      UNARY(MEMOF,{expr_loc = Convert_env.get_loc env; expr_node = var })
    | _ -> var

let global_reference env n =
  let (is_extern_c, typ) = Convert_env.get_global_var env n in
  let name = if is_extern_c then n.decl_name
    else
      let n, t = Convert_env.typedef_normalize env n TStandard in
      Mangling.mangle n t None in
       let var = VARIABLE name in
       match typ with
       | LVReference _ | RVReference _ ->
         UNARY(MEMOF,{expr_loc = Convert_env.get_loc env; expr_node = var })
       | _ -> var

let convert_variable env = function
  | Local({ decl_name = "__func__" }) -> 
    CONSTANT(CONST_STRING (Convert_env.get_current_func_name env))
  | Local n ->  local_reference env n.decl_name
  | Global n -> global_reference env n
  | FunctionParameter n -> local_reference env n
  | CodePointer (n,signature,kind,is_extern_c,tm) ->
    let cname =
      if is_extern_c then n.decl_name
      else
        let n, tm = Convert_env.typedef_normalize env n tm in
        let signature = Convert_env.signature_normalize env signature in
        Mangling.mangle n tm (Some (kind,signature))
    in VARIABLE cname

let convert_binary kind assgn e1 e2 =
  match kind,assgn with
    | BOPlus, AKRValue -> BINARY(ADD,e1,e2)
    | BOPlus, AKAssign -> BINARY(ADD_ASSIGN,e1,e2)
    | BOMinus, AKRValue -> BINARY(SUB,e1,e2)
    | BOMinus, AKAssign -> BINARY(SUB_ASSIGN,e1,e2)
    | BOLess, AKRValue -> BINARY(LT,e1,e2)
    | BOLessOrEqual, AKRValue -> BINARY(LE, e1, e2)
    | BOEqual, AKRValue -> BINARY(EQ,e1,e2)
    | BODifferent, AKRValue -> BINARY(NE,e1,e2)
    | BOGreaterOrEqual, AKRValue -> BINARY(GE,e1,e2)
    | BOGreater, AKRValue -> BINARY(GT,e1,e2)
    | BOTimes, AKRValue -> BINARY(MUL,e1,e2)
    | BOTimes, AKAssign -> BINARY(MUL_ASSIGN,e1,e2)
    | BODivide, AKRValue -> BINARY(DIV,e1,e2)
    | BODivide, AKAssign -> BINARY(DIV_ASSIGN,e1,e2)
    | BOModulo, AKRValue -> BINARY(MOD,e1,e2)
    | BOModulo, AKAssign -> BINARY(MOD_ASSIGN,e1,e2)
    | BOBitOr, AKRValue -> BINARY(BOR,e1,e2)
    | BOBitOr, AKAssign -> BINARY(BOR_ASSIGN,e1,e2)
    | BOBitAnd, AKRValue -> BINARY(BAND,e1,e2)
    | BOBitAnd, AKAssign -> BINARY(BAND_ASSIGN,e1,e2)
    | BOBitExclusiveOr, AKRValue -> BINARY(XOR,e1,e2)
    | BOBitExclusiveOr, AKAssign -> BINARY(XOR_ASSIGN,e1,e2)
    | BOLeftShift, AKRValue -> BINARY(SHL,e1,e2)
    | BOLeftShift, AKAssign -> BINARY(SHL_ASSIGN,e1,e2)
    | BORightShift, AKRValue -> BINARY(SHR,e1,e2)
    | BORightShift, AKAssign -> BINARY(SHR_ASSIGN,e1,e2)
    | BOLogicalAnd, AKRValue -> BINARY(AND,e1,e2)
    | BOLogicalOr, AKRValue -> BINARY(OR,e1,e2)
    | BOComma, AKRValue -> COMMA [e1;e2]
    | _, AKAssign -> 
        Frama_Clang_option.fatal
          "Binary operator is not supposed to have an assign kind"

let is_bin_assign = function
  | ADD | SUB | MUL | DIV | MOD | AND | OR | BAND | BOR | XOR
  | SHL | SHR | EQ | NE | LT | GT | LE | GE
    -> false
  | ASSIGN | ADD_ASSIGN | SUB_ASSIGN | MUL_ASSIGN | DIV_ASSIGN | MOD_ASSIGN
  | BAND_ASSIGN | BOR_ASSIGN | XOR_ASSIGN | SHL_ASSIGN | SHR_ASSIGN
    -> true

let is_unary_assign = function
  | MINUS | PLUS | NOT | BNOT | MEMOF | ADDROF -> false
  | PREINCR | PREDECR | POSINCR | POSDECR -> true

let rec make_addrof e =
  match e.expr_node with
    | VARIABLE _ | INDEX _ | MEMBEROF _ | MEMBEROFPTR _ -> 
       { e with expr_node = UNARY(ADDROF,e) }
    (* I think this is handled well by cabs2cil. *)
    | QUESTION _ -> { e with expr_node = UNARY(ADDROF,e) }
    | CAST(_,SINGLE_INIT e) -> make_addrof e
    | PAREN e -> make_addrof e
    | BINARY(a, e1, _) when is_bin_assign a ->
        { e with expr_node = COMMA [ e; make_addrof e1 ] }
    | UNARY(MEMOF, e) -> e
    | UNARY(a, e1) when is_unary_assign a ->
        { e with expr_node = COMMA [ e; make_addrof e1 ] }
    | COMMA l ->
        (match List.rev l with
           | [] ->
               Frama_Clang_option.fatal
                 "Trying to take the address of an empty expression"
           | a::l -> 
               { e with expr_node = COMMA (List.rev ((make_addrof a) :: l))})
    | NOTHING | UNARY _ | LABELADDR _ | BINARY _ | CALL _ | CONSTANT _
    | EXPR_SIZEOF _ | TYPE_SIZEOF _ | EXPR_ALIGNOF _ | TYPE_ALIGNOF _
    | GNU_BODY _ | EXPR_PATTERN _ | CAST _ ->
      Frama_Clang_option.fatal
        "Cannot take the address of a non-lval expression"

let is_builtin_va_list = function
  | Named({ decl_name = "__builtin_va_list"}, _) -> true
  | _ -> false

let rec convert_ref env typ expr =
  match typ, expr with
  (* special case for va_list. *)
  | (LVReference (PDataPointer ty)
    | RVReference (PDataPointer ty)), e
    when is_builtin_va_list ty.plain_type -> e
  | (LVReference _ | RVReference _), { expr_node = UNARY(MEMOF,e) } -> e
  | (LVReference _ | RVReference _), { expr_node = CAST(_,SINGLE_INIT e) } ->
      make_addrof e
  | (LVReference _ | RVReference _), _ -> make_addrof expr
  | Named (ty, _), _
      -> if Convert_env.has_typedef env ty
         then convert_ref env (Convert_env.get_typedef env ty).plain_type expr
         else expr
  | _, _ -> expr

let convert_reference_parameters env variadic prms args =
  let convert_ref typ arg = convert_ref env typ.plain_type arg in
  let rec convert = function
    | [], [] -> []
    | [], args when variadic -> args
    | prm::prms, arg::args -> convert_ref prm arg :: convert (prms,args)
    | _ ->
      Convert_env.fatal env
        "Wrong number of arguments in function call (expected %d, got %d)"
        (List.length prms) (List.length args)
  in
  convert (prms, args)

let mk_expr env node = { expr_loc = Convert_env.get_loc env; expr_node = node }

let rec is_constructor_call e = 
 match e.econtent with
   | Static_call (_, _, FKConstructor _, _, _,_) -> true
   | Unary_operator(UOCastNoEffect _,e) -> is_constructor_call e
   | _ -> false

let rec extract_constructor_call e =
  match e.econtent with
    | Static_call (name, sigtype, (FKConstructor _ as kind), args, tn, _) ->
      (kind, name, tn, sigtype, args)
    | Unary_operator(UOCastNoEffect _,e) -> extract_constructor_call e
    | _ -> Frama_Clang_option.fatal "Not a constructor"

let add_fc_destructor_attr env typ attrs =
  let loc = Convert_env.get_loc env in
  match (Convert_env.qual_type_normalize env typ).plain_type with
  | Struct (n,args) when Convert_env.has_destructor env (n,args) ->
    let name =
      Mangling.mangle
        (Cxx_utils.meth_name n args ("~" ^ n.decl_name))
        TStandard
        (Some
           (FKDestructor true,
            { result = Cxx_utils.unqual_type Void;
              parameter = [];
              variadic = false;
            }))
    in
    ("__declspec",
     [{ expr_loc = loc;
        expr_node =
          CALL
            ({ expr_loc = loc;
               expr_node = VARIABLE Cabs2cil.frama_c_destructor },
             [{ expr_loc = loc;
                expr_node =
                  UNARY(
                    ADDROF,
                    { expr_loc = loc;
                      expr_node =
                        CONSTANT (CONST_STRING name)})}])}]) :: attrs
  | _ -> attrs

let rm_fc_destructor_attr attrs =
  List.filter
    (fun (name,content) ->
       name <> "__declspec" ||
       (match content with
        | [ { expr_node = CALL ({ expr_node = VARIABLE n }, _)}] ->
          n <> Cabs2cil.frama_c_destructor
        | _ -> true))
    attrs

let add_temporary (* env *) e =
  if is_constructor_call e then
    let (kind, cons,tc,signature,args) = extract_constructor_call e in
(*  let typ =
      Convert_env.class_type_from_qualifications env cons.prequalification
    in
    let rt = { qualifier = []; plain_type = typ } in
    let name = Convert_env.temp_name env "__temp" in
    let args =
      { eloc = e.eloc; econtent = Variable(Local (empty_qual name)) } :: args
    in *)
    let e =
      { e with econtent
            = Static_call(cons,signature,kind,args,tc,false)
      }
    in e
    (* { e with econtent = Temporary(name,rt,Single_init e, false) } *)
  else e

(* some temporary might be needed. Note that their list is returned reverted.
   convert_full_expr takes care of putting it in the good order. *)
let shift_pointer_var_counter = ref 0
let virtual_var_counter = ref 0
let shift_object_counter = ref 0

let find_loc_list f l =
  match l with
    | [] -> Cil_datatype.Location.unknown
    | [ s ] -> f s
    | s :: l ->
        let (beg_loc,_) = f s in
        let rec aux = function
          | [] -> assert false
          | [ s ] ->
              let (_,end_loc) = f s in
              (beg_loc, end_loc)
          | _::l -> aux l
        in
        aux l

let find_loc_stmt =
  function
    | Nop l -> l
    | Return (l,_) -> l
    | Expression (l,_) -> l
    | VirtualExpression (l,_) -> l
    | Ghost_block(l,_) -> l
    | Block (l,_) -> l
    | Condition (l,_,_,_) -> l
    | Label(l,_) -> l
    | Goto(l,_) -> l
    | Switch(l,_,_) -> l
    | VarDecl(l,_,_,_) -> l
    | Break l -> l
    | Continue l -> l
    | While(l,_,_,_) -> l
    | DoWhile(l,_,_,_) -> l
    | For(l,_,_,_,_,_) -> l
    | Code_annot(l,_) -> l
    | TryCatch(l,_,_) -> l
    | GccInlineAsm(l,_,_,_) -> l

let find_loc_list_stmt = find_loc_list find_loc_stmt

let find_loc_case_stmt =
  function
    | Case(_,l) | Default l -> find_loc_list_stmt l

let find_loc_case_stmt_list = find_loc_list find_loc_case_stmt

let add_temp _env stmts (dec, init) =
  let stmts =
    match init with
      | None -> stmts
      | Some init -> init :: stmts
  in
  match dec with
    | None -> stmts
    | Some dec ->
      { stmt_ghost = false; stmt_node = DEFINITION dec } :: stmts

let add_temp_update env acc (dec,inits) =
  (* We have three possibilities:
     - no definition and a statement
     - a single init expression in the definition and no init statement
     - no init expression and an init statement *)
  match dec with
    | None ->
      (match inits with
        | None -> acc
        | Some s -> s :: acc)
    | Some (DECDEF(_,(_,[(name,_,_,_),inite]),_)) ->
      let stmt =
        match inite, inits with
          | SINGLE_INIT e, None ->
            { stmt_ghost = false;
              stmt_node =
                COMPUTATION(
                  mk_expr env (BINARY(ASSIGN, mk_expr env (VARIABLE name), e)),
                  Convert_env.get_loc env)}
          | NO_INIT, Some s -> s
          | _ ->
            Convert_env.fatal env "Unexpected initialization for a temporary"
      in
      stmt :: acc
    | _ -> Convert_env.fatal env "Unexpected definition for a temporary"

let mk_compound_init env lv typ init =
  let loc = Convert_env.get_loc env in
  let rec aux acc lv typ init =
    match init with
    | SINGLE_INIT def ->
      COMPUTATION (
        { expr_loc = def.expr_loc; expr_node = BINARY(ASSIGN,lv,def)},
        def.expr_loc)
      :: acc
    | NO_INIT -> NOP loc :: acc
    | COMPOUND_INIT l ->
      (match typ.plain_type with
       | Array { subtype } ->
         let rec aux_array acc i l =
           match l with
           | [] -> acc
           | (what,init) :: l ->
             (* translation scheme uses that for now. *)
             assert (what = NEXT_INIT);
             let lv =
               { expr_loc = lv.expr_loc;
                 expr_node =
                   INDEX(
                     lv,
                     { expr_loc = lv.expr_loc;
                       expr_node = CONSTANT (CONST_INT (string_of_int i))})}
             in
             aux_array (aux acc lv subtype init) (i+1) l
         in
         aux_array acc 0 l
       | Struct (name, tk) ->
         let rec aux_struct acc lfields linit =
           match lfields, linit with
           | _,[] -> acc
           | [],_ -> 
             Convert_env.fatal
               env "Too many initializers for class %a" 
               Fclang_datatype.Qualified_name.pretty (name,tk)
           | (fname,ftype)::lfields, (what,i)::linit ->
             assert (what = NEXT_INIT);
             let lv = {expr_loc=lv.expr_loc; expr_node=MEMBEROF(lv,fname)} in
             aux_struct (aux acc lv ftype i) lfields linit
         in
         aux_struct acc (Convert_env.get_struct env (name,tk)) l
       | _ -> Convert_env.fatal env "Compound init on a scalar type")
  in
  let init_stmts = aux [] lv typ init in
  match init_stmts with
  | [] -> assert false (* at least one initialization is supposed to occur. *)
  | [ single_init ] -> single_init
  | l ->
    let l = List.rev_map (fun x -> { stmt_ghost = false; stmt_node = x }) l in
    make_block env l

let computation_or_nop loc exp =
  match exp.expr_node with
  | NOTHING -> NOP loc
  | _ -> COMPUTATION(exp,loc)

let preserved_returned_object tmps e =
  match e.expr_node with
  | VARIABLE n ->
    let transf (d,s as res) =
      match d with
      | Some (DECDEF(spec,(t,l),loc)) ->
        let change_name ((n',decl,attrs,loc),init as res) =
          if n <> n' then res
          else
            ((n,decl,rm_fc_destructor_attr attrs,loc),init)
        in
        let l' = List.map change_name l in
        Some (DECDEF(spec,(t,l'),loc)),s
      | _ -> res
    in
    List.map transf tmps
  | _ -> tmps

let rec convert_base_type env spec decl typ does_remove_virtual =
  match typ with
  | Void -> spec_type Tvoid :: spec, decl
  | Int IBool -> spec_type Tbool :: spec, decl
  | Int (IChar_u | IChar_s | IChar) -> spec_type Tchar :: spec, decl
  | Int IUChar -> (List.map spec_type [Tunsigned; Tchar]) @ spec, decl
  | Int ISChar -> (List.map spec_type [Tsigned; Tchar ]) @ spec, decl
    (* TODO: intKindForSize returns a type of exactly 16 bits. There is no
       function for providing an ikind of at least 16 bits yet. This should
       be added to Cil. Indeed, it could theoretically be possible that 
       intKindForSize 2 fails while there exist types of a strictly
       greater size. *)
  | Int IChar16 -> spec_of_ikind (Cil.intKindForSize 2 true) @ spec, decl
  | Int IChar32 -> spec_of_ikind (Cil.intKindForSize 4 true) @ spec, decl
  | Int (IWChar_u | IWChar_s | IWUChar | IWSChar | IWChar ) ->
      let wchar_name = { prequalification=[];decl_name="fc_wchar_t"} in
      let base =
        if Convert_env.has_typedef env wchar_name then begin
          let rep = (Convert_env.get_typedef env wchar_name).plain_type in
          match rep with
            | Named (_,is_extern_c) ->
                let name =
                  if is_extern_c then wchar_name.decl_name
                  else Mangling.mangle wchar_name TStandard None
                in
                [ SpecType (Tnamed name) ]
            | Int _ ->
              let spec,_ =
                convert_base_type env [] decl rep does_remove_virtual
              in
              spec
            | _ ->
                Frama_Clang_option.fatal
                  "wchar_t should be linked to a named or integral type"
        end else
          spec_of_ikind Cil.theMachine.Cil.wcharKind
      in
      base @ spec, decl
  | Int IInt -> spec_type Tint :: spec, decl
  | Int IShort -> spec_type Tshort :: spec, decl
  | Int IUShort -> (List.map spec_type [Tunsigned; Tshort ]) @ spec, decl
  | Int IUInt -> (List.map spec_type [Tunsigned; Tint]) @ spec, decl
  | Int ILong -> spec_type Tlong :: spec, decl
  | Int IULong -> (List.map spec_type [Tunsigned; Tlong]) @ spec,decl
  | Int ILongLong -> (List.map spec_type [Tlong; Tlong]) @ spec,decl
  | Int IULongLong -> (List.map spec_type [Tunsigned; Tlong; Tlong]) @ spec,decl
  | Float FFloat -> spec_type Tfloat :: spec, decl
  | Float FDouble -> spec_type Tdouble :: spec, decl
  | Float FLongDouble -> (List.map spec_type [Tlong; Tdouble]) @ spec, decl
  | Enum e ->
    let body_name, t = Convert_env.typedef_normalize env e.body TStandard in
    spec_type (Tenum(Mangling.mangle body_name t None,None,[]))::spec, decl
  | Struct (s,t) ->
    let name =
      if Convert_env.is_extern_c_aggregate env s t then s.decl_name
      else
        let s, t = Convert_env.typedef_normalize env s t in
        Mangling.mangle s t None
    in
    spec_type (Tstruct (name, None, [])) :: spec, decl
  | Union (s,t) ->
    let name =
      if Convert_env.is_extern_c_aggregate env s t then s.decl_name
      else
        let s, t = Convert_env.typedef_normalize env s t in
        Mangling.mangle s t None
    in
    spec_type (Tunion (name, None, [])) :: spec, decl
  | Pointer (PDataPointer t) ->
    let attrs = List.map cv_to_attr spec in
    let decl d = decl (protect_ptr_type attrs d) in
    convert_type env decl t does_remove_virtual
  | LVReference (PDataPointer t) | RVReference(PDataPointer t)->
    let attrs = List.map cv_to_attr spec in
    convert_type
      env (fun d -> decl (protect_ptr_type attrs d)) t does_remove_virtual
  | Pointer (PFunctionPointer s) ->
    let args = convert_signature env s.parameter does_remove_virtual in
    let rt, rt_decl = convert_specifiers env s.result does_remove_virtual in
    let attrs = List.map cv_to_attr spec in
    rt,
    (fun d ->
      rt_decl
        (PROTO (decl (protect_ptr_type attrs d), args,s.variadic)))
  | LVReference (PFunctionPointer s)
  | RVReference (PFunctionPointer s) ->
    let args = convert_signature env s.parameter does_remove_virtual in
    let rt, rt_decl = convert_specifiers env s.result does_remove_virtual in
    let attrs= List.map cv_to_attr spec in
    rt,
    (fun d ->
       rt_decl (PROTO (decl (protect_ptr_type attrs d),args,s.variadic)))
  | Pointer(PStandardMethodPointer _)
  | LVReference (PStandardMethodPointer _) 
  | RVReference (PStandardMethodPointer _) ->
      Frama_Clang_option.not_yet_implemented "pointer to member"
  | Pointer(PVirtualMethodPointer _)
  | LVReference (PVirtualMethodPointer _)
  | RVReference (PVirtualMethodPointer _) ->
      Frama_Clang_option.not_yet_implemented "pointer to member"
  | Array a ->
    let is_array_attribute = function
      | SpecCV _ -> true
      | _ -> false
    in
    let attrs = Extlib.filter_map is_array_attribute cv_to_attr spec in
      convert_type
        env
        (fun d ->
          let dim =
            Extlib.may_map
              (fun e ->
                 let _,_,ce = convert_expr env [] e does_remove_virtual in
                 ce.expr_node)
              ~dft:NOTHING
              a.dimension
          in
          let exp =
            { expr_loc = Cil_datatype.Location.unknown; expr_node = dim }
          in
          decl (protect_array_type attrs exp d))
        a.subtype
        does_remove_virtual
  | Named (name, is_extern_c_name) ->
    let cname =
      if Cxx_utils.is_builtin_qual_type name then name.decl_name
      else if is_extern_c_name
      then name.decl_name
      else
        let name, t = Convert_env.typedef_normalize env name TStandard in
        Mangling.mangle name t None
    in
    spec_type (Tnamed cname)::spec, decl

and convert_type env decl t does_remove_virtual =
  let spec = List.map convert_cv t.qualifier in
  convert_base_type env spec decl t.plain_type does_remove_virtual

and convert_signature env l does_remove_virtual =
  match l with
    | [] ->
      (* in C++, an empty list is strictly equivalent to (void), i.e. no 
         argument at all. In C, a prototype with no argument means that the
         arguments are not specified, so that a subsequent declaration could
         provides one or more arguments. We thus normalize that to (void) for
         the C translation.
       *)
      [ [SpecType Tvoid],("",JUSTBASE,[],Convert_env.get_loc env) ]
    | _ ->
      List.map (convert_anonymous_decl env does_remove_virtual) l

and convert_specifiers env t does_remove_virtual =
  let spec = List.map convert_cv t.qualifier in
  let rt, decl =
    convert_base_type env spec (fun d -> d) t.plain_type does_remove_virtual
  in
  rt, decl

and convert_anonymous_decl env does_remove_virtual t =
  let typ, decl = convert_specifiers env t does_remove_virtual in
  (typ, ("",decl JUSTBASE,[],Cil_datatype.Location.unknown))

and convert_decl env arg does_remove_virtual =
  let typ,decl = convert_specifiers env arg.arg_type does_remove_virtual in
  (typ, (arg.arg_name, decl JUSTBASE, [], arg.arg_loc))

and convert_constant env c does_remove_virtual = match c with
  | IntCst (t,_,v) ->
    if ((Int64.compare v Int64.zero) >= 0)
    then CONSTANT (make_integral_constant_kind t v)
    else UNARY(MINUS, {expr_loc = Convert_env.get_loc env;
        expr_node = (CONSTANT (make_integral_constant_kind t (Int64.neg v)))})
  | FloatCst(_,v) -> CONSTANT(CONST_FLOAT v)
  | EnumCst(n,e,_) ->
    let n, t = Convert_env.typedef_normalize env n TStandard in
    let name = Mangling.mangle n t None in
    let body_name, t' = Convert_env.typedef_normalize env e.body TStandard in
    let enum = Mangling.mangle body_name t' None in
    (* C++ enum constant are of type Enum, while C treat them as integers.
       This is not an issue for most purposes, except when it comes to
       handle exceptions: catching enum e is not the same as catching int x.
     *)
    CAST (([SpecType (Tenum (enum,None,[]))], JUSTBASE),
          SINGLE_INIT(
            { expr_loc = Convert_env.get_loc env;
              expr_node = VARIABLE name }))
  | TypeCst (TCCSizeOf, t) ->
    let bt,decl = convert_base_type env [] id t does_remove_virtual in
       (TYPE_SIZEOF (bt,decl JUSTBASE))
  | TypeCst (TCCAlignOf, t) ->
    let bt,decl = convert_base_type env [] id t does_remove_virtual in
       (TYPE_ALIGNOF (bt,decl JUSTBASE))

and convert_unary env kind arg does_remove_virtual =
  match kind with
    | UOCastNoEffect _ -> arg.expr_node
    (* Not a real cast, merely a compilation's artifact *)
    | UOCastInteger(t,_)
    | UOCastEnum(t,_) | UOCastFloat(t,_) | UOCastC t ->
      let rt, decl = convert_specifiers env t does_remove_virtual in
      CAST((rt, decl JUSTBASE), (SINGLE_INIT arg))
    | UOPostInc -> UNARY(POSINCR,arg)
    | UOPostDec -> UNARY(POSDECR,arg)
    | UOPreInc  -> UNARY(PREINCR,arg)
    | UOPreDec  -> UNARY(PREDECR,arg)
    | UOOpposite -> UNARY(MINUS,arg)
    | UOBitNegate -> UNARY(BNOT,arg)
    | UOLogicalNegate -> UNARY(NOT,arg)

(* drop_temp is true when the resulting value is not considered further,
   i.e. the expression is evaluated only for its side effect. In this setting,
   temporaries will be translated as NOTHING, since computations occur in the
   tmps instructions and not the returned expression.
*)
and convert_expr_node ?(drop_temp=false) env tmps e does_remove_virtual =
  let create_this_access e origin_type tmps env is_reference noeffect =
    if (noeffect)
    then e, tmps, env
    else begin
      let derived_name, td
        = if not is_reference
          then Convert_env.get_class_name_from_pointer env
            origin_type.plain_type
          else let derived_name, td, _ =
            Convert_env.get_class_name_from_reference env
              origin_type.plain_type
          in derived_name, td
      in let var_name = "__cast_tmp_"
        ^ (string_of_int !shift_pointer_var_counter)
      in begin
        incr shift_pointer_var_counter;
        let env = Convert_env.add_local_var env var_name
          (Pointer ( PDataPointer
            { qualifier = [];
              plain_type = Struct (derived_name,td) }))
        in let derived_name, td
          = Convert_env.typedef_normalize env derived_name td
        in let init = if not is_reference
          then e
          else { expr_loc = e.expr_loc ;
                 expr_node = UNARY(ADDROF, e)}
        in let tmps = (Some (DECDEF(
           None,
           ([SpecType
             (Tstruct
               (Mangling.mangle derived_name td None, None, []))],
            [(var_name, PTR([], JUSTBASE),[], e.expr_loc),
             SINGLE_INIT(init)]),
           e.expr_loc)), None) :: tmps
        in { expr_loc = e.expr_loc;
             expr_node = VARIABLE var_name
           }, tmps, env
       end
    end
  in
  let create_table_access access tmps env noeffect = if (noeffect)
    then access, tmps, env
    else begin
      let table_access = "__virtual_tmp_"
        ^ (string_of_int !virtual_var_counter)
      in let qual_vmt_content_name
        = { prequalification=[]; decl_name="_frama_c_vmt_content" }
      in
      incr virtual_var_counter;
      let env = Convert_env.add_local_var env table_access 
        (Pointer (PDataPointer { qualifier=[];
          plain_type = Struct (qual_vmt_content_name, TStandard) }))
      in let qual_vmt_content_name, _
        = Convert_env.typedef_normalize env qual_vmt_content_name TStandard
      in let tmp_decl = DECDEF( None,
       ([SpecType (Tstruct (Mangling.mangle
           qual_vmt_content_name TStandard None, None, []))],
         [(table_access,PTR( [], JUSTBASE),[],access.expr_loc),
           SINGLE_INIT (access)]),
       access.expr_loc)
      in (mk_expr env (VARIABLE table_access)),
         (Some tmp_decl, None) :: tmps, env
    end
  in
  let create_shift_object tmps env loc =
    let var_name = "__shift_object"
      ^ (string_of_int !shift_object_counter)
    in begin
      incr shift_object_counter;
      let env = Convert_env.add_local_var env var_name
        (Pointer ( PDataPointer
          { qualifier = [];
            plain_type = Int IInt }))
      in let tmps = (Some (DECDEF(
         None,
         ([SpecType Tint],
          [(var_name, JUSTBASE, [], loc),
           SINGLE_INIT { expr_loc = loc;
                         expr_node = CONSTANT (CONST_INT "0") }]),
         loc)), None) :: tmps
      in { expr_loc = loc;
           expr_node = VARIABLE var_name
         }, tmps, env
     end
  in
  let env, tmps, node =
    match e with
      | Constant c -> env, tmps, (convert_constant env c does_remove_virtual)
      | String s -> env, tmps, CONSTANT (CONST_STRING s)
      | Variable v -> env, tmps, convert_variable env v
      | Malloc(t) ->
          let bt,decl = convert_base_type env [] id t does_remove_virtual in
          env,
          tmps,
          CALL(mk_expr env (VARIABLE "malloc"),
               [mk_expr env (TYPE_SIZEOF (bt,decl JUSTBASE))])
      | MallocArray(t,s) ->
          let bt, decl = convert_base_type env [] id t does_remove_virtual in
          let env, tmps, size = convert_expr env tmps s does_remove_virtual in
          env,
          tmps,
          CALL(mk_expr env (VARIABLE "malloc"),
               [mk_expr env
                   (BINARY(MUL,
                           mk_expr env (TYPE_SIZEOF(bt,decl JUSTBASE)),size))])
      | Free e | FreeArray e ->
          let env, tmps, arg = convert_expr env tmps e does_remove_virtual in
          env, tmps, CALL(mk_expr env (VARIABLE "free"),[arg])
      | Assign(x,e) when is_constructor_call e ->
        let kind, name, tn, sigtype, args = extract_constructor_call e in
        let e = 
          { e with
            econtent =
              Static_call(name, sigtype, kind,
                { x with econtent = Address x}::args, tn, false)}
        in
        let env, tmps, e = convert_expr env tmps e does_remove_virtual in
        env, tmps, e.expr_node
      | Assign(x,e) ->
          let env, tmps, lv = convert_expr env tmps x does_remove_virtual in
          let env, tmps, rv = convert_expr env tmps e does_remove_virtual in
          env, tmps, BINARY(ASSIGN,lv,rv)
      | AssignReference(x,e) ->
          let env, tmps, lv = convert_expr env tmps x does_remove_virtual in
          let env, tmps, rv =
            convert_expr env tmps
              ({ eloc = e.eloc; econtent = Address(e) }) does_remove_virtual
          in
          env, tmps, BINARY(ASSIGN,lv,rv)
      | Unary_operator(k,e) ->
          let env, tmps, e = convert_expr env tmps e does_remove_virtual in
          env, tmps, convert_unary env k e does_remove_virtual
      | Binary_operator(k,a,e1,e2) ->
          let env, tmps, e1 = convert_expr env tmps e1 does_remove_virtual in
          let env, tmps, e2 = convert_expr env tmps e2 does_remove_virtual in
          env, tmps, convert_binary k a e1 e2
      | Dereference e ->
          let env, tmps, e = convert_expr env tmps e does_remove_virtual in
          env, tmps, UNARY(MEMOF,e)
      | Address e ->
          let env, tmps, e = convert_expr env tmps e does_remove_virtual in
          env, tmps, (make_addrof e).expr_node
      | PointerCast(target,base,e) ->
          let env, tmps, e = convert_expr env tmps e does_remove_virtual in
          let rt, decl = convert_specifiers env target does_remove_virtual in
            (match base with
              | RPKPointer ->
                  env, tmps, CAST((rt,decl JUSTBASE),SINGLE_INIT e)
              | RPKReference ->
                  env, tmps, UNARY(MEMOF,
                    { expr_loc = e.expr_loc ;
                      expr_node =  CAST((rt,decl JUSTBASE),SINGLE_INIT
                      ( { expr_loc = e.expr_loc ;
                          expr_node = UNARY(ADDROF,e)}))})
              | RPKStaticBasePointer ->
                let (base_class_name, t) =
                  Convert_env.get_class_name_from_pointer env target.plain_type
                in let (base_class_name, t) =
                  Convert_env.typedef_normalize env base_class_name t
                in
                let cname = "_frama_c_" ^ Mangling.mangle base_class_name t None
                in
                env, tmps,
                CAST
                  ((rt,decl JUSTBASE),
                   SINGLE_INIT
                     { expr_loc = e.expr_loc ;
                       expr_node =
                         UNARY(
                             ADDROF,
                             { expr_loc = e.expr_loc ;
                               expr_node = MEMBEROFPTR (e,cname) })})
              | RPKStaticBaseReference ->
                let (base_class_name, t, _) =
                  Convert_env.get_class_name_from_reference
                    env target.plain_type
                in let (base_class_name, t) =
                  Convert_env.typedef_normalize env base_class_name t
                in
                let cname = "_frama_c_" ^ Mangling.mangle base_class_name t None
                in
                env, tmps,
                UNARY
                  (MEMOF,
                   { expr_loc = e.expr_loc ;
                      expr_node =
                        CAST
                          ((rt,decl JUSTBASE),
                           SINGLE_INIT
                             { expr_loc = e.expr_loc ;
                               expr_node =
                                 UNARY
                                   (ADDROF,
                                    { expr_loc = e.expr_loc ;
                                      expr_node =
                                        MEMBEROF(e, cname)})})})
              | RPKVirtualBasePointer(base_index, origin_type,
                    noeffect) ->
                let derived_name, td
                  = Convert_env.get_class_name_from_pointer env
                      origin_type.plain_type
                in let var_name = "__cast_tmp_"
                  ^ (string_of_int !shift_pointer_var_counter)
                in incr shift_pointer_var_counter;
                let env = Convert_env.add_local_var env var_name
                  (Pointer ( PDataPointer
                    { qualifier = [];
                      plain_type = Struct (derived_name,td) })) in
                let tmps = (Some (DECDEF(
                   None,
                   ([SpecType
                        (Tstruct
                           (Mangling.mangle derived_name td None, None, []))],
                    [(var_name, PTR([], JUSTBASE),[], e.expr_loc),
                     SINGLE_INIT(e)]),
                   e.expr_loc)), None) :: tmps in
                let qual_vmt_name
                  = { prequalification=[]; decl_name="_frama_c_vmt" }
                in let this_access, tmps, env =
                  create_this_access e origin_type tmps env false noeffect
                in let qual_vmt_name, _
                  = Convert_env.typedef_normalize env qual_vmt_name TStandard
                in let table_access_def =
                  mk_expr env (BINARY
                    (ADD, (mk_expr env (MEMBEROFPTR
                      ((mk_expr env (UNARY(MEMOF, mk_expr env (CAST(
                        ([SpecType (Tstruct (Mangling.mangle qual_vmt_name
                            TStandard None, None, []))],
                         PTR ([], PTR ([], JUSTBASE))),
                        SINGLE_INIT(this_access)))))), "table"))),
                      (mk_expr env (CONSTANT (CONST_INT
                        (Int64.to_string base_index))))))
                in let access, tmps, env =
                  create_table_access table_access_def tmps env noeffect
                in env, tmps, CAST((rt,decl JUSTBASE),
                  SINGLE_INIT(mk_expr env
                    (BINARY
                       (ADD (* to keep positive shift by default
                               - virtual base classes are after *),
                        (* could add downcast with shift *)
                        (mk_expr
                           env
                           (CAST
                              (([SpecType Tchar], PTR ([], JUSTBASE)),
                               SINGLE_INIT this_access))),
                        (mk_expr
                           env
                           (MEMBEROFPTR (access, "shift_this")))))))
              | RPKVirtualBaseReference(base_index, origin_type) ->
                let this_access, tmps, env =
                  create_this_access e origin_type tmps env
                    true (* is_reference *) false (* noeffect *)
                in let qual_vmt_name
                  = { prequalification=[]; decl_name="_frama_c_vmt" }
                in let loc = Convert_env.get_loc env
                in let qual_vmt_name,_
                  = Convert_env.typedef_normalize env qual_vmt_name TStandard
                in let table_access_def = 
                  mk_expr env (BINARY
                    (ADD, (mk_expr env (MEMBEROFPTR
                      ((mk_expr env (UNARY(MEMOF, mk_expr env (CAST(
                        ([SpecType (Tstruct (Mangling.mangle qual_vmt_name
                            TStandard None, None, []))],
                         PTR ([], PTR ([], JUSTBASE))),
                        SINGLE_INIT(
                        this_access)))))), "table"))),
                      (mk_expr env (CONSTANT (CONST_INT
                        (Int64.to_string base_index))))))
                in let access, tmps, env =
                  create_table_access table_access_def tmps env false
                      (* noeffect *)
                in env, tmps, UNARY(MEMOF,
                  { expr_loc = loc;
                    expr_node = CAST((rt,decl JUSTBASE),
                    SINGLE_INIT(mk_expr env
                     (BINARY
                       (ADD (* to keep positive shift by default
                               - virtual base classes are after *),
                        (* could add downcast with shift *)
                        (mk_expr
                           env
                           (CAST
                              (([SpecType Tchar], PTR ([], JUSTBASE)),
                               SINGLE_INIT this_access))),
                        (mk_expr
                           env
                           (MEMBEROFPTR (access, "shift_this")))))))})
              | RPKDynamicPointer (origin_type, pvmt) ->
                let shift_object, tmps, env
                  = create_shift_object tmps env e.expr_loc
                in let origin_name, td = Convert_env
                  .get_class_name_from_pointer env origin_type.plain_type
                in let origin_qualification = 
                  match td with
                    | TStandard -> QStructOrClass(origin_name.decl_name)
                    | TTemplateInstance params ->
                        QTemplateInstance(origin_name.decl_name, params)
                in
                let env, tmps, fst_arg =
                  convert_expr_node env tmps
                    (Address
                       { eloc = Convert_env.get_loc env;
                         econtent =
                           (Variable
                              (Global 
                                 { prequalification = List.append
                                     origin_name.prequalification
                                     [origin_qualification];
                                   decl_name = "_frama_c_rtti_name_info" }))})
                    does_remove_virtual
                in
                let env, tmps, snd_arg =
                  convert_expr env tmps pvmt does_remove_virtual
                in
                let target_name, ttd =
                  Convert_env .get_class_name_from_pointer env target.plain_type
                in
                let target_qualification = 
                  match ttd with
                    | TStandard -> QStructOrClass(target_name.decl_name)
                    | TTemplateInstance params ->
                        QTemplateInstance(target_name.decl_name, params)
                in
                let env, tmps, thd_arg =
                  convert_expr_node env tmps
                    (Address
                       { eloc = Convert_env.get_loc env;
                         econtent =
                           Variable
                             (Global 
                                { prequalification = List.append
                                    target_name.prequalification
                                    [target_qualification];
                                  decl_name = "_frama_c_rtti_name_info" })})
                    does_remove_virtual
                in let fourth_arg = mk_expr env (UNARY(ADDROF, shift_object))
                in let args = [fst_arg; snd_arg; thd_arg; fourth_arg]
                in let dynamic_cast_name =
                  (Mangling.mangle { prequalification = [];
                                     decl_name = "_frama_c_find_dynamic_cast"}
                    TStandard (Some (FKFunction,
                      { result = { qualifier = []; plain_type = Int(IInt) };
                        parameter =
                          [ { qualifier = [];
                              plain_type = (Pointer(PDataPointer
                                { qualifier = [];
                                  plain_type = Struct(
                                    { prequalification = [];
                                      decl_name = "_frama_c_rtti_name_info_node"
                                    }, TStandard)}))};
                            { qualifier = [];
                              plain_type = (Pointer(PDataPointer
                                { qualifier = [];
                                  plain_type = Struct(
                                    { prequalification = [];
                                      decl_name = "_frama_c_vmt"
                                    }, TStandard)}))};
                            { qualifier = [];
                              plain_type = (Pointer(PDataPointer
                                { qualifier = [];
                                  plain_type = Struct(
                                    { prequalification = [];
                                      decl_name = "_frama_c_rtti_name_info_node"
                                    }, TStandard)}))};
                            { qualifier = [];
                              plain_type = (Pointer(PDataPointer
                                { qualifier = [];
                                  plain_type = Int(IInt)}))}];
                        variadic = false})))
                in
                  env, tmps, QUESTION(
                    { expr_loc = e.expr_loc;
                      expr_node = CALL(mk_expr env
                        (VARIABLE dynamic_cast_name),args)
                    },
                    { expr_loc = e.expr_loc;
                      expr_node = CAST((rt,decl JUSTBASE),
                      SINGLE_INIT
                        { expr_loc = e.expr_loc;
                          expr_node = BINARY(ADD,
                            { expr_loc = e.expr_loc;
                              expr_node = CAST(([SpecType Tvoid],
                                  PTR ([], JUSTBASE)), SINGLE_INIT e)
                            },
                            shift_object)
                        })
                    },
                    { expr_loc = e.expr_loc;
                      expr_node = CAST((rt,decl JUSTBASE),
                      SINGLE_INIT { expr_loc = e.expr_loc;
                                    expr_node = CONSTANT (CONST_INT "0") })})
              | RPKDynamicReference (_, _) ->
                  env, tmps, UNARY(MEMOF,
                    { expr_loc = e.expr_loc ;
                      expr_node =  CAST((rt,decl JUSTBASE),SINGLE_INIT
                      ( { expr_loc = e.expr_loc ;
                          expr_node = UNARY(ADDROF,e)}))}))
      | ShiftPointerCast(target,base,kind,e) ->
          let env, tmps, e = convert_expr env tmps e does_remove_virtual in
          let rt, decl = convert_specifiers env target does_remove_virtual in
          let derived_aux e derived_name td base_name tb =
            let var_name = "__cast_tmp_"
              ^ (string_of_int !shift_pointer_var_counter) in
            let env = Convert_env.add_local_var env var_name
                        (Pointer ( PDataPointer
                           { qualifier = [];
                             plain_type = Struct (derived_name,td) })) in
            let derived_name,td
              = Convert_env.typedef_normalize env derived_name td in
            let tmps =
              (Some 
                 (DECDEF(
                   None,
                   ([SpecType
                        (Tstruct
                           (Mangling.mangle derived_name td None, None, []))],
                    [(var_name, PTR([], JUSTBASE),[], e.expr_loc),
                     SINGLE_INIT
                       { expr_loc = e.expr_loc;
                         expr_node =
                           CAST 
                             (([SpecType
                                   (Tstruct (
                                     Mangling.mangle
                                       derived_name td None, None, []))],
                               PTR([], JUSTBASE)),
                              SINGLE_INIT
                                { expr_loc = e.expr_loc;
                                  expr_node = CONSTANT (CONST_INT "0")
                                })}]),
                   e.expr_loc)), None) :: tmps
            in
            let var_tmp =
              { expr_loc = e.expr_loc; expr_node = VARIABLE var_name }
            in
            let type_char_ptr e =
               { expr_loc = e.expr_loc ;
                 expr_node =
                   CAST(([SpecType Tchar], PTR ([], JUSTBASE)),SINGLE_INIT e) }
            in
            let base_name,tb
              = Convert_env.typedef_normalize env base_name tb in
            let shift = BINARY(SUB,
              (type_char_ptr var_tmp),
              (type_char_ptr
                { expr_loc = e.expr_loc;
                  expr_node = UNARY(ADDROF,
                  { expr_loc = e.expr_loc;
                    expr_node = MEMBEROFPTR (var_tmp,
                      "_frama_c_" ^ Mangling.mangle base_name tb None)
                  })
                }))
            in
            env, tmps, CAST((rt,decl JUSTBASE),SINGLE_INIT
              { expr_loc = e.expr_loc ;
                expr_node = BINARY(ADD, (type_char_ptr e), 
                  { expr_loc = e.expr_loc ;
                    expr_node = shift
                  })})
          in
            (match kind with
              | RPKPointer | RPKStaticBasePointer ->
                let derived_name, td =
                  Convert_env.get_class_name_from_pointer env target.plain_type
                in 
                let base_name, tb =
                  Convert_env.get_class_name_from_pointer env base.plain_type
                in
                incr shift_pointer_var_counter;
                derived_aux e derived_name td base_name tb
              | RPKReference | RPKStaticBaseReference ->
                let derived_name, td, _ =
                  Convert_env.get_class_name_from_reference
                    env target.plain_type
                in
                let base_name, tb, is_base_reference =
                  Convert_env.get_class_name_from_reference env base.plain_type
                in
                incr shift_pointer_var_counter;
                if is_base_reference
                then
                  let env, tmps, expr =
                    derived_aux e derived_name td base_name tb
                  in
                  env, tmps,
                  UNARY(MEMOF, { expr_loc = e.expr_loc; expr_node = expr })
                else
                  let env, tmps, expr =
                    derived_aux
                      { expr_loc = e.expr_loc; expr_node = UNARY(ADDROF, e) }
                      derived_name td base_name tb
                  in
                  env, tmps,
                  UNARY(MEMOF, { expr_loc = e.expr_loc; expr_node = expr })
              | RPKVirtualBasePointer(_, _, _)
              | RPKVirtualBaseReference(_,_) ->
                Frama_Clang_option.fatal
                  "virtual base class cannot be converted in derived class"
              | RPKDynamicPointer(_, _)
              | RPKDynamicReference(_, _) ->
                Frama_Clang_option.fatal
                  "dynamic_cast class is not support in shift_pointer class")
      | FieldAccess(e,f) ->
          let env, tmps, e = convert_expr env tmps e does_remove_virtual in
          env, tmps, MEMBEROF(e,f)
      | ArrayAccess(a,i) ->
          let env, tmps, a = convert_expr env tmps a does_remove_virtual in
          let env, tmps, i = convert_expr env tmps i does_remove_virtual in
          env, tmps, INDEX(a,i)
      | Conditional(econd,etrue, efalse) ->
          let env, tmps, econd =
            convert_expr env tmps econd does_remove_virtual
          in
          let env, tmps, etrue =
            convert_expr env tmps etrue does_remove_virtual
          in
          let env, tmps, efalse =
            convert_expr env tmps efalse does_remove_virtual
          in
          env, tmps, QUESTION(econd, etrue, efalse)
      | Static_call(name, signature, kind, args, t, is_extern_c) ->
        let cname =
          if is_extern_c then name.decl_name
          else
            let name, t = Convert_env.typedef_normalize env name t in
            let signature = Convert_env.signature_normalize env signature in
            Mangling.mangle name t (Some (kind,signature))
        in
        let class_type cv =
          let class_type = Convert_env.class_type_from_qualifications
            env name.prequalification
          in
          { qualifier = [];
            plain_type =
              Pointer
                (PDataPointer { qualifier = cv; plain_type = class_type })}
        in
        let signature =
          match kind with
            | FKFunction  -> signature
            | FKDestructor _ | FKConstructor _ ->
              { signature with
                parameter = (class_type []) :: signature.parameter }
            | FKMethod cv | FKCastMethodOperator (cv,_) ->
              { signature with
                parameter = (class_type cv) :: signature.parameter }
        in
        let env, tmps, args =
          convert_list_expr env tmps args does_remove_virtual
        in
        let prm = remove_void signature.parameter in
        let args = convert_reference_parameters env signature.variadic prm args
        in
        env, tmps, CALL(mk_expr env (VARIABLE cname),args)
      | Virtual_call(name,signature,_(*kind*),this,args,method_index, shift,
          shiftPvmt)
        ->
        let env, tmps, this = convert_expr env tmps this does_remove_virtual in
        let env, tmps, args =
          convert_list_expr env tmps args does_remove_virtual
        in
        let prm = remove_void signature.parameter in
        let args = convert_reference_parameters env false prm args in
        let var_name = "__virtual_tmp_" ^ (string_of_int !virtual_var_counter)
        in let qual_vmt_content_name
          = { prequalification=[]; decl_name="_frama_c_vmt_content" } in
        let qual_vmt_name
          = { prequalification=[]; decl_name="_frama_c_vmt" } in
        incr virtual_var_counter;
        let env = Convert_env.add_local_var env var_name 
          (Pointer (PDataPointer { qualifier=[];
            plain_type = Struct (qual_vmt_content_name, TStandard) }))
        in
        let loc = Convert_env.get_loc env in
        let shift_expr this_expr
          { base = base_class_name;
            templated_kind = t;
            is_virtual = _ (* is_base_virtual *);
            vmt_position = _ (* index *);
          }
          = let base_class_name, t
            = Convert_env.typedef_normalize env base_class_name t
          in { expr_loc = this_expr.expr_loc ;
                expr_node = UNARY(ADDROF,
                  { expr_loc = this_expr.expr_loc ;
                    expr_node = MEMBEROFPTR(this_expr,
                      ("_frama_c_"
                        ^ Mangling.mangle base_class_name t None))})}
        in let thisPvmt = begin
          match shiftPvmt with
             | [] -> this
             | _ -> (mk_expr env (CAST (([SpecType Tchar], PTR ([], JUSTBASE)),
                SINGLE_INIT (List.fold_left shift_expr this shiftPvmt)))) 
        end in
        let qual_vmt_content_name, _
          = Convert_env.typedef_normalize env qual_vmt_content_name TStandard in
        let qual_vmt_name, _
          = Convert_env.typedef_normalize env qual_vmt_name TStandard in
        let tmp_decl = DECDEF( None,
          ([SpecType (Tstruct (Mangling.mangle
              qual_vmt_content_name TStandard None, None, []))],
            [(var_name,PTR( [], JUSTBASE),[],loc),
              SINGLE_INIT (mk_expr env (BINARY
                (ADD, (mk_expr env (MEMBEROFPTR
                  ((mk_expr env (UNARY(MEMOF, mk_expr env (CAST(
                    ([SpecType (Tstruct (Mangling.mangle qual_vmt_name
                        TStandard None, None, []))],
                     PTR ([], PTR ([], JUSTBASE))),
                    SINGLE_INIT(thisPvmt)))))), "table"))),
                  (mk_expr env (CONSTANT (CONST_INT
                    (Int64.to_string method_index)))))))]), loc) in 
        let class_type = Convert_env.class_type_from_qualifications
            env name.prequalification in
        let proto_args =
          convert_signature
            env
            ({ qualifier = [];
               plain_type =
                 Pointer
                   (PDataPointer
                      { qualifier = [];
                        plain_type = class_type})} :: signature.parameter)
            does_remove_virtual
        in
        let proto_rt, proto_rt_decl =
          convert_specifiers env signature.result does_remove_virtual
        in
        let proto_spec = List.map convert_cv signature.result.qualifier in
        let class_name, tc =
          let rev_name = (List.rev name.prequalification) in
          let declared_name, tc =
          begin 
            match (List.hd rev_name) with
              | QStructOrClass name -> (name, TStandard)
              | QTemplateInstance (name, l) -> name, (TTemplateInstance l)
              | _ -> Convert_env.fatal env
                "Unexpected qualification for virtual call"
          end in
          ({ prequalification = (List.rev (List.tl rev_name));
            decl_name = declared_name
           }, tc)
        in let class_name, tc
          = Convert_env.typedef_normalize env class_name tc
        in
        env,
        (Some tmp_decl, None)::tmps,
        CALL(
            mk_expr
              env
              (UNARY
                 (MEMOF,
                  mk_expr
                    env
                    (CAST
                       ((proto_rt,
                         proto_rt_decl
                           (PROTO
                              ((PTR(List.map cv_to_attr proto_spec,JUSTBASE)),
                               proto_args,false))),
                        SINGLE_INIT
                          (mk_expr
                             env
                             (MEMBEROFPTR ((mk_expr env (VARIABLE var_name)),
                                           "method_ptr"))))))),
            (mk_expr
               env
               (CAST 
                  (([SpecType
                       (Tstruct
                          (Mangling.mangle class_name tc None, None, []))],
                    PTR ([], JUSTBASE)),
                   SINGLE_INIT(
                       mk_expr
                         env
                         (BINARY
                            (SUB (* to keep positive shift by default
                                    - concrete classes are before *),
                             (* could add downcast with shift *)
                             (mk_expr
                                env
                                (CAST
                                   (([SpecType Tchar], PTR ([], JUSTBASE)),
                                    SINGLE_INIT
                                      (List.fold_left shift_expr this shift)))),
                             (mk_expr
                                env
                                (MEMBEROFPTR ((mk_expr env (VARIABLE var_name)),
                                              "shift_this"))))))))) :: args)
      | Dynamic_call(_kind,ptr,args) ->
        (* NB: might be slightly different for virtual method *)
        let signature = Convert_env.get_dynamic_signature env ptr.econtent in
        let env, tmps, f = convert_expr env tmps ptr does_remove_virtual in
        let env, tmps, args =
          convert_list_expr env tmps args does_remove_virtual
        in
        let prm = remove_void signature.parameter in
        let args = convert_reference_parameters env signature.variadic prm args
        in env, tmps, CALL(f,args)
      | Temporary(name, ctyp, init_exp, force) ->
        let env = Convert_env.add_local_var env name ctyp.plain_type in
        let typ, decl = convert_specifiers env ctyp does_remove_virtual in
        let attrs = add_fc_destructor_attr env ctyp [] in
        let res = if drop_temp then NOTHING else VARIABLE name in
        let rec aux exp =
          match exp.econtent with
            | Static_call (n,t,(FKConstructor _ as kind),args,tm,_) ->
              (* clang seems to insert the temporary in the list
                 of arguments, but only randomly... 
                 Note that the parameter's list never contain the receiver
                 argument.
               *)
              let args = 
                if List.length args = List.length t.parameter then
                  { eloc = exp.eloc;
                    econtent = (Address { eloc = exp.eloc;
                      econtent = (Variable (Local (empty_qual name)))})}
                  :: args
                else args
              in
              let env, tmps, def =
                convert_constr_expr
                  env tmps n kind tm t args does_remove_virtual
              in
              let stmt =
                { stmt_ghost = false;
                  stmt_node = COMPUTATION (def, def.expr_loc) }
              in
              let loc = Convert_env.get_loc env in
              let tmp_decl =
                DECDEF(None,(typ,[(name,decl JUSTBASE,attrs,loc), NO_INIT]),loc)
              in
              env, (Some tmp_decl, Some stmt)::tmps, res
            | Unary_operator(UOCastNoEffect _,e) -> aux e
            | Assign({ econtent = Variable(Local{decl_name = n})}, e)
                when n = name ->
              aux e
            | AssignReference({ econtent = Variable(Local{decl_name = n})}, e)
                when n = name ->
              aux e
            | _ ->
              let env, tmps, def =
                convert_expr env tmps exp does_remove_virtual
              in
              (match def.expr_node, force with
                | VARIABLE _ as v, false -> env, tmps, v
                (* no need to waste a temporary here *)
                | _ ->
                  let loc = Convert_env.get_loc env in
                  let tmp_decl =
                    DECDEF(
                      None,
                      (typ,[(name,decl JUSTBASE,attrs,loc), SINGLE_INIT def]),
                      loc)
                  in
                  env, (Some tmp_decl, None)::tmps, res)
        in
        (match init_exp with
          | Single_init exp -> aux exp
          | _ ->
            let loc = Convert_env.get_loc env in
            let var = Local { prequalification = []; decl_name = name } in
            let env, tmps', init, _ =
              convert_initializer env ctyp var init_exp does_remove_virtual
            in
            let tmps = tmps' @ tmps in
            let tmp_decl =
              DECDEF(
                None,
                (typ,[(name,decl JUSTBASE,attrs,loc), init]), loc)
            in
            env, (Some tmp_decl, None)::tmps, res)
      | VAArg(e,t) ->
        let env, tmps', e = convert_expr env tmps e does_remove_virtual in
        let typ, decl = convert_specifiers env t does_remove_virtual in
        (* implement the ugly comment in cabs.ml *)
        let builtin = mk_expr env (VARIABLE "__builtin_va_arg") in
        let typ_expr = mk_expr env (TYPE_SIZEOF (typ, decl JUSTBASE)) in
        env, tmps' @ tmps, CALL (builtin,[e;typ_expr])
      | Throw None ->
        let stmt =
          { stmt_ghost = false;
            stmt_node = THROW (None, Convert_env.get_loc env) }
        in
        env, (None, Some stmt)::tmps, NOTHING
      | Throw (Some e) ->
        let loc = Convert_env.get_loc env in
        let env, tmps, e = convert_expr env tmps e does_remove_virtual in
        let tmps = preserved_returned_object tmps e in
        let stmt =
          { stmt_ghost = false;
            stmt_node = THROW(Some e, loc) }
        in
        env, (None, Some stmt)::tmps, NOTHING
      | GnuBody l ->
         env,
         tmps,
         GNU_BODY
           {blabels = [];
            battrs = [];
            bstmts =
              List.rev
                (snd
                   (List.fold_left
                      (fun (env,l) s ->
                       let stmt, env =
                         convert_statement env s does_remove_virtual
                       in
                       (env, stmt::l))
                      (env, [])
                      l))}
  in
  env, tmps, mk_expr env node

and convert_expr ?drop_temp env tmps e =
  let env = Convert_env.set_loc env e.eloc in
  convert_expr_node ?drop_temp env tmps e.econtent

and convert_list_expr env tmps l does_remove_virtual =
  let env, tmps, args =
    List.fold_left
      (fun (env, tmps, acc) arg ->
        let arg = add_temporary (* env *) arg in
        let env, tmps, arg = convert_expr env tmps arg does_remove_virtual in
        env, tmps, arg::acc)
      (env, tmps, []) l
  in env, tmps, List.rev args

and convert_constr_expr env tmps n kind tc t args does_remove_virtual =
  let n', tc'
    = Convert_env.typedef_normalize env n tc in
  let t' = Convert_env.signature_normalize env t in
  let cname = Mangling.mangle n' tc' (Some (kind,t')) in
  let class_type =
    Convert_env.class_type_from_qualifications env n.prequalification
  in
  let this_type = 
    { qualifier = [];
      plain_type =
        Pointer (PDataPointer { qualifier = []; plain_type = class_type})}
  in
  let signature = { t with parameter = this_type :: t.parameter} in
  let env, tmps, args = convert_list_expr env tmps args does_remove_virtual in
  let prm = remove_void signature.parameter in
  let args = convert_reference_parameters env t.variadic prm args in
  env,
  tmps,
  mk_expr env (CALL(mk_expr env (VARIABLE cname),args))

and convert_full_constr_expr env n kind tc t args =
  convert_constr_expr env [] n kind tc t args

(* list of temp declaration is built backwards. Don't forget to revert it
   somewhere. *)
and convert_full_expr ?drop_temp env e = convert_expr ?drop_temp env [] e

and combine_trunc l1 l2 =
  match l1,l2 with
    | [], _ | _,[] -> []
    | x::xs, y::ys -> (x,y) :: combine_trunc xs ys

and find_type_list env typ l =
  let rec aux typ =
    match typ with
      | Void | Int _ | Enum _ | Float _ | Pointer _ 
      | LVReference _ | RVReference _ ->
        Convert_env.fatal env "Using compound initialization for a scalar value"
      | Array typ ->
        List.map (fun x -> (typ.subtype, x)) l
      | Struct (s,ts) ->
        let fields = Convert_env.get_struct env (s,ts) in
        combine_trunc (List.map snd fields) l
      | Union _ -> (* handled by Union_init, not Compound_init *)
        Convert_env.fatal env "Using compound initialization for an union"
      (* builtins are always scalar for now. *)
      | Named (ty,_) when Cxx_utils.is_builtin_qual_type ty ->
        Convert_env.fatal env
          "Using compound initialization for a builtin value"
      | Named (ty,_) -> aux (Convert_env.get_typedef env ty).plain_type
  in aux typ

and convert_initializer env typ var init_exp does_remove_virtual =
  let init_var_counter = ref 0 in
  let equal v1 v2 =
    match v1,v2 with
      | Local v1, Local v2
        -> Fclang_datatype.Qualified_name.equal (v1,TStandard) (v2,TStandard)
      | Global v1, Global v2
        -> Fclang_datatype.Qualified_name.equal (v1,TStandard) (v2,TStandard)
      | _ -> false
  in
  let rec aux env typ var = function
    | Single_init init ->
      (match init.econtent with
        | Static_call(n,t, (FKConstructor _ as kind),args,tm,_) ->
          let env, tmps, def =
            convert_full_constr_expr env n kind tm t args does_remove_virtual
          in
          env,
          tmps, NO_INIT,
          Some ({ stmt_ghost = false;
                  stmt_node = COMPUTATION (def, def.expr_loc)})
        | Assign({ econtent = Variable v}, e)
            when equal v var ->
          aux env typ var (Single_init e)
        | AssignReference({ econtent = Variable v}, e)
            when equal v var ->
          aux env typ var (Single_init e)
        | _ ->
          let env, tmps, def = convert_full_expr env init does_remove_virtual in
          let def = convert_ref env typ.plain_type def in
          env, tmps, SINGLE_INIT def, None)
    | Compound_init l ->
      let typed_l = find_type_list env typ.plain_type l in
      let env, tmps, init =
        List.fold_left (convert_one_init NEXT_INIT) (env, [],[]) typed_l
      in
      let init =
        match init with
        | [] -> NO_INIT
        | _ -> COMPOUND_INIT (List.rev init)
      in 
      env, tmps, init, None
    | Array_init(idx, init) ->
      (match Convert_env.qual_type_normalize env typ with
       | { plain_type = Array { dimension } } ->
         (match dimension with
            | Some ({ econtent = Constant (IntCst (_,_,v))}) ->
              let idx = idx.decl_name in
              let loc = Convert_env.get_loc env in
              let lenv =
                Convert_env.add_local_var env idx (Int IULong)
              in
              let tmp =
                DECDEF(
                  None,
                  ([SpecType Tunsigned; SpecType Tlong],
                   [(idx, JUSTBASE, [], loc),
                    SINGLE_INIT(
                      {expr_loc = loc;
                       expr_node = CONSTANT (CONST_INT "0UL")})]),
                  loc)
              in
              let eidx = { expr_loc = loc; expr_node = VARIABLE idx } in
              let end_test =
                { expr_loc = loc;
                  expr_node =
                    BINARY(
                      LT,
                      eidx,
                      { expr_loc = loc;
                        expr_node =
                          CONSTANT (CONST_INT (Int64.to_string v)) }) }
              in
              let increment =
                { expr_loc = loc; expr_node = UNARY(POSINCR, eidx)}
              in
              let env, tmps, body =
                convert_full_expr ~drop_temp:true lenv init does_remove_virtual in
              let stmt_node = computation_or_nop loc body in
              let body = { stmt_ghost = false; stmt_node } in
              env, tmps, NO_INIT,
              Some(
                { stmt_ghost = false;
                  stmt_node = FOR([],FC_DECL tmp,end_test,increment,body,loc) })
            | _ ->
              Convert_env.fatal env
                "Implicit array initialization \
                 over a array of non-constant size"
         )
       | _ ->
         Convert_env.fatal env
           "Implicit array initialization on a non-array type")
    | Union_init(field,typ,init) ->
      let env, tmps, init =
        convert_one_init
          (INFIELD_INIT (field,NEXT_INIT)) (env, [],[]) (typ, init)
      in
      assert (List.length init = 1);
      env, tmps, COMPOUND_INIT init, None
  and convert_one_init what (env, tmps, acc) (ty,i) =
    let tmp = "__init_tmp_" ^ (string_of_int !init_var_counter) in
    let var = Local { prequalification = []; decl_name = tmp } in
    let env, new_tmps, def, cons = aux env ty var i in
    let tmps = new_tmps @ tmps in
    match i, cons with
      | _, None -> env, tmps, (what,def)::acc
      | Single_init e, Some _ ->
        incr init_var_counter;
        let spec, decl = convert_specifiers env ty does_remove_virtual in
        let attrs = add_fc_destructor_attr env ty [] in
        let decl =
          DECDEF
            (None,
             (spec, [(tmp, decl JUSTBASE,attrs,e.eloc), NO_INIT]), e.eloc)
        in
        env,
        (Some decl, cons)::tmps,
        (what,
         SINGLE_INIT { expr_loc = e.eloc; expr_node = VARIABLE tmp })
        :: acc
      | _, Some _ ->
        Convert_env.fatal env "unexpected initializer"
  in
  aux env typ var init_exp

and convert_init_statement env init does_remove_virtual =
  let loc = Convert_env.get_loc env in
  match init with
  | INop ->
    FC_EXP { expr_loc = loc; expr_node = NOTHING },[], env
  | IExpression e ->
    let env, tmps, e = convert_full_expr ~drop_temp:true env e does_remove_virtual in
    FC_EXP e, List.fold_left (add_temp env) [] tmps, env
  | IVarDecl init_declarator_list ->
    let env, tmps, l, def, base =
      List.fold_right
        (fun {id_name; init_type=typ; init_val} (env,tmps,l,def, base) ->
          let (base', decl) = convert_specifiers env typ does_remove_virtual in
          let base =
            match base with
             | None -> Some base'
             | Some b ->
               if b = base' then base
               else
                 Convert_env.fatal env
                   "list of declaration with different base types"
          in
          let attrs = add_fc_destructor_attr env typ [] in
          let env = Convert_env.add_local_var env id_name typ.plain_type in
          match init_val with
          | None ->
	    let init = (id_name, decl JUSTBASE,[],loc),NO_INIT in
	    (env, tmps, init::l, def, base)
          | Some init ->
            let var = Local { prequalification = []; decl_name = id_name } in
            let env,tmps',init, def' =
              convert_initializer env typ var init does_remove_virtual
            in
	    let def = match def' with
	      | None -> def
	      | Some stmt -> stmt::def
	    in
	    let init = (id_name, decl JUSTBASE,attrs,loc),init in
	    env, tmps'@tmps, init::l, def, base)
	init_declarator_list
	(env, [], [], [], None)
    in
    let tmps = List.rev tmps in
    let l = List.rev l in
    if l = [] then
      Convert_env.fatal env "Empty list of local variable declarations";
    let base = Extlib.the base in
    let decl = DECDEF (None,(base,l),loc) in
    match def with
    | [] -> FC_DECL decl, List.fold_left (add_temp env) [] tmps, env
    | l ->
      FC_EXP { expr_loc = loc; expr_node = NOTHING },
      List.fold_left (add_temp env) l ((Some decl,None)::tmps),
      env

and convert_condition env cond does_remove_virtual =
  let loc = Convert_env.get_loc env in
  match cond with
  | CondExpression e ->
    let env, res, e = convert_full_expr env e does_remove_virtual in res, e, env
  | CondVarDecl(name, typ, init) ->
    let attrs = add_fc_destructor_attr env typ [] in
    let env = Convert_env.add_local_var env name typ.plain_type in
    let (base, decl) = convert_specifiers env typ does_remove_virtual in
    let res = { expr_loc = loc; expr_node = VARIABLE name } in
    let var = Local { prequalification = []; decl_name = name } in
    let env, tmps, init, e =
      convert_initializer env typ var init does_remove_virtual
    in
    let decl =
      DECDEF(None,(base,[(name, decl JUSTBASE,attrs,loc),init]),loc)
    in
    (Some decl, e) :: tmps, res, env

and convert_incr_statement env incr does_remove_virtual =
  let loc = Convert_env.get_loc env in
  match incr with
    | CNop -> env, { expr_loc = loc; expr_node = NOTHING }
    | CExpression e -> 
      let env, tmps, e = convert_full_expr ~drop_temp:true env e does_remove_virtual in
      let stmt_node = computation_or_nop loc e in
      match tmps with
      | [] -> env, e
      | _ ->
        let stmt = { stmt_ghost = false; stmt_node } in
        let stmts = List.fold_left (add_temp env) [stmt] tmps in
        env, 
        { expr_loc = loc;
          expr_node = GNU_BODY { blabels=[]; battrs=[]; bstmts = stmts }
        }

and convert_statement env st does_remove_virtual =
  let rec add_temps tmps stmt =
    match tmps with
    | [] -> stmt
    | (def,init) :: tl ->
      let loc = Convert_env.get_loc env in
      let stmt =
        match init with
          | None -> stmt
          | Some s ->
            SEQUENCE(s, { stmt_ghost = false; stmt_node = stmt},loc)
      in
      let stmt =
        match def with
          | None -> stmt
          | Some def ->
            let def = { stmt_ghost = false; stmt_node = DEFINITION def } in
            let stmt = { stmt_ghost = false; stmt_node = stmt } in
            SEQUENCE(def,stmt,loc)
      in
      add_temps tl stmt
  in
  let add_temps_update tmps stmt =
    match tmps with
    | [] -> stmt
    | _ ->
      let block = List.fold_left (add_temp_update env) [] tmps in
      make_block_stmt env (stmt::block)
  in
  let raw, env = match st with
    | Nop l -> NOP l, env
    | Code_annot (l,annot) ->
      let env = Convert_env.set_loc env l in
      let c_annot = Convert_acsl.convert_code_annot env annot in
      CODE_ANNOT(c_annot,l), env
    | Return (l,Some e) ->
      let e = add_temporary (* env *) e in
      let env, tmps, e = convert_full_expr env e does_remove_virtual in
      let e = convert_ref env (Convert_env.get_current_return_type env) e in
      let tmps = preserved_returned_object tmps e in
      add_temps tmps (RETURN(e,l)), env
    | Return(l, None) -> RETURN({expr_loc = l; expr_node = NOTHING},l), env
    | Expression (loc,e) ->
      let env = Convert_env.set_loc env loc in
      let env, tmps, e = convert_full_expr ~drop_temp:true env e does_remove_virtual in
      add_temps tmps (COMPUTATION(e,loc)), env
    | VirtualExpression (loc,e) ->
      if does_remove_virtual
      then NOP loc, env
      else
        let env = Convert_env.set_loc env loc in
        let env, tmps, e = convert_full_expr env e does_remove_virtual in
        add_temps tmps (COMPUTATION(e,loc)), env
    | Ghost_block(loc,stmts) ->
      let env = Convert_env.set_loc env loc in
      let b = convert_block env stmts does_remove_virtual in
      List.iter (fun s -> s.stmt_ghost <- true) b.bstmts;
      BLOCK(b,loc,loc), env
    | Block(loc,stmts) ->
      let env = Convert_env.set_loc env loc in
      BLOCK(convert_block env stmts does_remove_virtual, loc, loc), env
    | Condition(loc,cond,true_action,false_action) ->
      let env = Convert_env.set_loc env loc in
      let tmps, cond, env = convert_condition env cond does_remove_virtual in
      let true_action
        = fst (convert_statement env true_action does_remove_virtual) in
      let false_action
        = fst (convert_statement env false_action does_remove_virtual) in
      add_temps tmps (IF(cond,true_action,false_action,loc)), env
    | Label(loc,lab) ->
      let env = Convert_env.set_loc env loc in
      let stmt = { stmt_ghost = false; stmt_node = NOP loc } in
      LABEL(lab,stmt,loc), env
    | Goto(loc,lab) ->
      let env = Convert_env.set_loc env loc in GOTO(lab,loc), env
    | Switch(loc,cond,cases) ->
      let env = Convert_env.set_loc env loc in
      let tmps, cond, env = convert_condition env cond does_remove_virtual in
      let cases_loc = find_loc_case_stmt_list cases in
      let cases, env =
        List.fold_left (convert_case_statement does_remove_virtual) ([],env) cases
      in let cases = { blabels = []; battrs = []; bstmts = List.rev cases } in
      let cases =
        { stmt_ghost = false; stmt_node = BLOCK(cases,cases_loc,cases_loc) }
      in
      add_temps tmps (SWITCH(cond,cases,loc)), env
    | VarDecl (loc, name, typ, init) ->
      let env = Convert_env.set_loc env loc in
      let (base, decl) = convert_specifiers env typ does_remove_virtual in
      let qual_name = { prequalification = []; decl_name = name } in
      let attrs = add_fc_destructor_attr env typ [] in
      let env =
        if List.mem Static typ.qualifier then
          let extern = Convert_env.is_extern_c env in
          let env = Convert_env.set_extern_c env true in
          let env = Convert_env.add_global_var env qual_name typ.plain_type in
          Convert_env.set_extern_c env extern
        else
          Convert_env.add_local_var env name typ.plain_type
      in
      let var = Local qual_name in
      let env, tmps, init, e =
        match init with
        | None -> env, [], NO_INIT, None
        | Some e -> convert_initializer env typ var e does_remove_virtual
      in
      let decl init =
        DECDEF(None, (base,[(name, decl JUSTBASE,attrs,loc),init]), loc)
      in
      let stmt =
        match tmps with
        | [] ->
          let def = DEFINITION (decl init) in
          (match e with
            | None -> def
            | Some stmt ->
              SEQUENCE ({ stmt_ghost = false; stmt_node = def}, stmt, loc))
        | _ ->
          (* We put all these declarations in a special block, 
             but the initial declaration itself need to stay out of it.
           *)
          let def = DEFINITION (decl NO_INIT) in
          let block =
            match e with
            | None ->
              let var_expr = { expr_loc = loc; expr_node = VARIABLE name } in
              mk_compound_init env var_expr typ init
            | Some s -> s.stmt_node
          in
          let block = add_temps tmps block in
          SEQUENCE ({ stmt_ghost = false; stmt_node = def },
                    { stmt_ghost = false; stmt_node = block }, loc)
      in
      stmt, env
    | Break l -> BREAK l, env
    | Continue l -> CONTINUE l, env
    | While (loc, cond, body, annot) ->
      let env = Convert_env.set_loc env loc in
      let env, tmps, cond = convert_full_expr env cond does_remove_virtual in
      let body,_ = convert_statement env body does_remove_virtual in
      let body = add_temps_update tmps body in
      let annot = List.map (Convert_acsl.convert_code_annot env) annot in
      add_temps tmps (WHILE(annot,cond,body,loc)), env
    | DoWhile (loc, cond, body,annot) ->
      let env = Convert_env.set_loc env loc in
      let env, tmps, cond = convert_full_expr env cond does_remove_virtual in
      let body, _ = convert_statement env body does_remove_virtual in
      let body = add_temps_update tmps body in
      let annot = List.map (Convert_acsl.convert_code_annot env) annot in
      add_temps tmps (DOWHILE(annot, cond, body, loc)), env
    | For(loc, init, cond, incr, body,annot) ->
      let env = Convert_env.set_loc env loc in
      let init, stmts, ienv =
        convert_init_statement env init does_remove_virtual
      in
      let ienv, tmps, cond = match cond with
        | None -> ienv, [], { expr_loc = loc ; expr_node = NOTHING }
        | Some cond -> convert_full_expr ienv cond does_remove_virtual
      in
      let ienv, incr = convert_incr_statement ienv incr does_remove_virtual in
      let body, _ = convert_statement ienv body does_remove_virtual in
      let body = add_temps_update tmps body in
      let annot = List.map (Convert_acsl.convert_code_annot ienv) annot in
      let loop = FOR(annot,init,cond,incr,body,loc) in
      let stmt =
        match stmts, tmps with
          [], [] -> loop
        | [], _ -> add_temps tmps loop
        | _,[] -> make_block env (stmts @ [{stmt_ghost = false; stmt_node = loop}])
        | _ ->
          let loop = { stmt_ghost = false; stmt_node = loop } in
          let block = List.fold_left (add_temp env) [loop] tmps in
          make_block env (stmts @ block)
      in
      stmt, env
    | TryCatch (loc, t, c) ->
      let env = Convert_env.set_loc env loc in
      let t, _ = convert_stmt_list env t does_remove_virtual in
      let t = List.rev t in
      let t = { stmt_ghost = false; stmt_node = make_block env t } in
      let c = List.fold_left (convert_catch_clause env does_remove_virtual) [] c
      in let c = List.rev c in
      TRY_CATCH(t,c,loc), env
    | GccInlineAsm(loc, qual, instr, details) ->
      let qual = List.map (fun x -> cv_to_attr (convert_cv x)) qual in
      let convert_asm_IO asm_io =
          List.map
            (fun {aIO_name=n;constraints=c;expr=e}->
             let (_,_,e)=convert_full_expr env e does_remove_virtual in (n,c,e))
            asm_io
      in
      let details = match details with
        | None -> None
        | Some {outputs=o; inputs=i; clobbers=c; ad_labels=l} ->
          Some 
            { aoutputs=convert_asm_IO o;
              ainputs=convert_asm_IO i;
              aclobbers=c; alabels=l }
      in
      ASM(qual, instr, details, loc), env
  in
  { stmt_ghost = false; stmt_node = raw }, env

and convert_catch_clause env does_remove_virtual acc { typed_arg; cbody } =
  let var, env =
    match typed_arg with
      | None -> None, env
      | Some v ->
        let qtype = Convert_env.qual_type_normalize env v.arg_type in
        Some (convert_decl env v does_remove_virtual),
        Convert_env.add_local_var env v.arg_name qtype.plain_type
  in
  let body, _ = convert_stmt_list env cbody does_remove_virtual in
  let body = List.rev body in
  let body = { stmt_ghost = false; stmt_node = make_block env body } in
  (var,body)::acc

and convert_case_statement does_remove_virtual (acc, env) case =
  match case with
    | Case(value,action) ->
        let loc = find_loc_list_stmt action in
        let value = convert_constant env value does_remove_virtual in
        let value = { expr_loc = loc; expr_node = value } in
        let action, env = convert_stmt_block env action does_remove_virtual in
        { stmt_ghost = false; stmt_node = CASE(value,action,loc) }::acc, env
    | Default(action) ->
        let loc = find_loc_list_stmt action in
        let action, env = convert_stmt_block env action does_remove_virtual in
        { stmt_ghost = false; stmt_node = DEFAULT(action,loc) }::acc, env

and convert_stmt_block env b does_remove_virtual =
  let loc = find_loc_list_stmt b in
  let stmts, env = convert_stmt_list env b does_remove_virtual in
  let stmt =
    match stmts with
      [] -> { stmt_ghost = false; stmt_node = NOP loc }
    | [ a ] -> a
    | l -> 
      let b = { blabels = []; battrs = []; bstmts = List.rev l } in
      { stmt_ghost = false; stmt_node = BLOCK(b,loc,loc) }
  in
  stmt, env

and convert_block env b does_remove_virtual =
  let stmts, _ = convert_stmt_list env b does_remove_virtual in
   { blabels = [];
     battrs = [];
     bstmts = List.rev stmts;
   }

(* Warning: convert_stmt_list gives back the translated list in reverse order*)
and convert_stmt_list env l does_remove_virtual =
  List.fold_left
    (fun (acc, env) stmt ->
      let (s,env) = convert_statement env stmt does_remove_virtual
      in s::acc, env) ([],env) l

let make_prototype loc env name rt args variadic does_remove_virtual =
  let rt, decl = convert_specifiers env rt does_remove_virtual in
  let args = 
    match args with
      | [] -> (* empty list in C++ always mean void, not unspecified *)
        [[SpecType Tvoid],("",JUSTBASE,[],loc)]
      | _ -> List.map (fun t -> convert_decl env t does_remove_virtual) args
  in
  (rt, (name,decl (PROTO(JUSTBASE,args,variadic)),[],loc))

let convert_enum_constant env ikind e =
  let loc = Convert_env.get_loc env in
  match e with
    | EnumCst(name,_,value) ->
        let name, t
          = Convert_env.typedef_normalize env name TStandard in
        let name = Mangling.mangle name t None in
        (* If the value is negative, create an expression
           with a positive value. *)
        let value =
          if value < Int64.zero then begin
            (* The negation of min_int may overflow. *)
            if value = Int64.min_int then
              Frama_Clang_option.abort
                ~current:true "Enum constant has value -2^63";
            UNARY(MINUS, 
                  mk_expr env
                    (CONSTANT
                       (make_integral_constant_kind ikind (Int64.neg value))))
          end else
            CONSTANT (make_integral_constant_kind ikind value)
        in
        (name,mk_expr env value,loc)
    | FloatCst _ | IntCst _ | TypeCst _ ->
        Frama_Clang_option.fatal "Enum constant has a wrong kind"

let convert_enum_constants env ikind l =
  List.map (convert_enum_constant env ikind) l

let convert_static_const env loc name ikind kind value does_remove_virtual =
  let env = Convert_env.set_loc env loc in
  let ghost = Convert_env.is_ghost env in
  let qualified_name = Convert_env.qualify env name in
  let name = 
    if Convert_env.is_extern_c env then name
    else
      let qualified_name, t
        = Convert_env.typedef_normalize env qualified_name TStandard in
      Mangling.mangle qualified_name t None
  in
  let plain = Int ikind in
  let spec, _ =
    convert_base_type env [SpecCV CV_CONST] id plain does_remove_virtual
  in
  let spec =
    match kind with
    | ICStaticConst -> SpecStorage STATIC :: spec
    | ICLiteral | ICExternConst -> spec
  in
  let init = { expr_loc = loc;
               expr_node = CONSTANT (CONST_INT (Int64.to_string value))}
  in
  let env = Convert_env.add_global_var env qualified_name plain in
  (ghost,
   DECDEF(None,(spec,[(name,JUSTBASE,[],loc),SINGLE_INIT init]),loc)), env
 
let remove_this_parameter kind args =
  match kind with
  | FKFunction -> args
  | FKMethod _ | FKCastMethodOperator _ | FKConstructor _ | FKDestructor _
    -> List.tl args

let convert_bitfield_info env l =
  mk_expr env (CONSTANT (CONST_INT (string_of_int l)))

let make_qualified_name env = function
  | Declaration string_name -> (Convert_env.qualify env string_name)
  | Implementation qual_name -> qual_name

let make_class_env env name kind tc is_extern_c =
  let qualified_name = make_qualified_name env name in
  let env = Convert_env.add_aggregate env qualified_name kind tc is_extern_c in
  match name with
    | Declaration string_name ->
      begin
      match tc with
      | TStandard -> Convert_env.add_namespace env (QStructOrClass string_name)
      | TTemplateInstance tc -> Convert_env.add_namespace env
          (QTemplateInstance (string_name, tc) )
      end
    | Implementation qual_name ->
      let subenv =  Convert_env.set_namespace env qualified_name in
      match tc with
      | TStandard -> Convert_env.add_namespace subenv
          (QStructOrClass qual_name.decl_name)
      | TTemplateInstance tc -> Convert_env.add_namespace subenv
          (QTemplateInstance (qual_name.decl_name, tc) )

let create_base_field_name env base tb =
  let base, tb = Convert_env.typedef_normalize env base tb in
  "_frama_c_" ^ (Mangling.mangle base tb None)

let make_class_decl env name tkind kind inherits fields body has_virtual =
  let n, t
    = Convert_env.typedef_normalize env name tkind in
  let class_name = Mangling.mangle n t None in
  let ghost = Convert_env.is_ghost env in
  let loc = Convert_env.get_loc env in
  let create_class_type base tb =
    { qualifier = []; plain_type = (Struct (base,tb)) }
  in
  let fields, bare_fields, new_env =
    match body with
      | None -> None, None, Convert_env.add_struct env (name, tkind) []
      | Some _ ->
        let rec add_base_class (base, tb) result = match result with
          | [] -> [(base, tb)]
          | x :: l -> if (Fclang_datatype.Qualified_name.equal x (base, tb))
            then x :: l
            else x :: (add_base_class (base, tb) l)
        in
        let rec append_bases l result = match l with
          | [] -> result
          | x :: l' -> (append_bases l' (add_base_class x result))
        in
        let virtual_base_classes =
          match inherits with
            | None -> []
            | Some inherits ->
                List.fold_left
                  (fun result inherits -> match inherits.is_virtual with
                    | VStandard -> 
                      append_bases (Class.get_virtual_base_classes
                        (inherits.base, inherits.templated_kind)) result
                    | VVirtual ->
                      add_base_class (inherits.base, inherits.templated_kind)
                        (append_bases (Class.get_virtual_base_classes
                            (inherits.base, inherits.templated_kind)) result))
                [] inherits
        in
        let virtual_inherited_fields = List.fold_left
          (fun result (base,tb) ->
            let n, t = Convert_env.typedef_normalize env base tb in
            let n' = 
              if Class.has_virtual_base_class (base,tb)
              then { n with decl_name = n.decl_name ^ "_frama_c_bare" }
              else n
            in (FIELD
               ([SpecType
                   (Tstruct
                      (Mangling.mangle n' t None,
                       None, []))],
                [(create_base_field_name env n t,
                  JUSTBASE, [], loc),
                 None]),
              (create_class_type base tb))
            ::result)
          [] virtual_base_classes
        in
        let has_virtual, inherited_fields =
          match inherits with
            | None -> has_virtual, []
            | Some inherits ->
                List.fold_left
                  (fun (has_virtual,result) inherits ->
                     let n,t =
                       Convert_env.typedef_normalize
                         env inherits.base inherits.templated_kind
                     in
                     let has_virtual =
                       has_virtual || Convert_env.struct_has_virtual env (n,t)
                     in
                     match inherits.is_virtual with
                     | VStandard -> 
                       let n, t = Convert_env.typedef_normalize
                           env inherits.base inherits.templated_kind
                       in
                       let n' = 
                         if Class.has_virtual_base_class
                             (inherits.base,inherits.templated_kind)
                         then { n with decl_name=n.decl_name ^ "_frama_c_bare" }
                         else n
                       in
                       has_virtual,
                       (FIELD
                         ([SpecType
                             (Tstruct
                                (Mangling.mangle n' t None,
                                 None, []))],
                          [(create_base_field_name env n t,
                            JUSTBASE, [], loc),
                           None]),
                        (create_class_type
                          inherits.base inherits.templated_kind))
                      :: result
                    | VVirtual -> has_virtual, result)
                (has_virtual, [])
                inherits
        in
        let create_field (cfs, fts) (cf,ft) =
          cf::cfs, match cf with
            | FIELD (_, name_init_list)
              -> (List.fold_left
                 (fun l ((name,_,_,_),_) -> (name, ft)::l) fts name_init_list)
            | _ -> fts
        in
        let cfields, fields_typ =
          List.fold_left create_field ([],[])
            (List.append virtual_inherited_fields
              (List.append fields inherited_fields))
        in
        let env = Convert_env.add_struct env (name,tkind) fields_typ in
        let env =
          if has_virtual then Convert_env.virtual_struct env (name,tkind)
          else env
        in
        let bare_cfields, env =
          match virtual_base_classes with
          | [] -> None, env
          | _ ->
             let bcfields, bfields_typ = 
               (List.fold_left create_field ([],[])
                 (List.append fields inherited_fields))
             in
             let env = Convert_env.add_aggregate env
                 { name with decl_name=name.decl_name ^ "_frama_c_bare" }
                 CClass tkind false
             in
             Some bcfields,
             Convert_env.add_struct env
               ({ name with decl_name=name.decl_name ^ "_frama_c_bare" },tkind)
               bfields_typ
        in
        Some cfields, bare_cfields, env
  in
  let type_decl =
    match kind with
      | CClass | CStruct -> Tstruct (class_name, fields,[])
      | CUnion -> Tunion(class_name, fields,[])
  in
  let bare_type_decl = match bare_fields with
    | None -> None
    | Some fields -> begin match kind with
      | CClass | CStruct -> Some (Tstruct (
        Mangling.mangle
        { n with decl_name = (n.decl_name ^ "_frama_c_bare") } t None,
        Some fields,[]))
      | CUnion -> None
    end
  in
  let bare_class_decl = match bare_type_decl with
    | None -> None
    | Some tdecl -> Some (ghost,DECDEF(None,([SpecType tdecl],[]),loc))
  in
  new_env, (ghost,DECDEF(None,([SpecType type_decl],[]),loc)), bare_class_decl

let is_implicit_func = function
  | CMethod(_,_,_,_,_,_,_,b,_,_,_,_) -> b
  | _ -> false

let extract_decl = function
  | CMethod(loc,name, kind,rt,args,variadic,_,_,tkind,has_further_definition,
      _,_) ->
    (* If any, the spec will still be ported by the original declaration. *)
    CMethod(loc,name,kind,rt,args,variadic,None,false,tkind,
            has_further_definition,None,None)
  | c -> c

let reorder_implicit l =
  let implicit = Extlib.filter_map is_implicit_func extract_decl l in
  implicit @ l

let iter_on_array ?(incr=true) env idx length mk_body =
  let loc = Convert_env.get_loc env in
  let idx_var = mk_expr env (VARIABLE idx) in
  let const_int i = CONSTANT(CONST_INT (string_of_int i)) in
  let init =
    if incr then const_int 0
    else BINARY(SUB, length,
		{expr_loc=loc;
		 expr_node=const_int 1})
  in
  let last = if incr then length.expr_node else const_int 0 in
  let last_test = if incr then LT else GE in
  let modif = if incr then POSINCR else POSDECR in
  let decl =
    DECDEF(
      None,(
        [SpecType Tunsigned],
        [(idx,JUSTBASE,[],loc),
         SINGLE_INIT (mk_expr env init)]),
      loc)
  in
  let body, def, env = mk_body env idx_var in
  { stmt_ghost = false;
    stmt_node =
      FOR([],FC_DECL decl,
          mk_expr env
            (BINARY(last_test,idx_var,mk_expr env last)),
          mk_expr env
            (UNARY(modif,idx_var)),
          body,loc) }, def, env

let create_stmt_set_vmt env qualification_class inherits =
  let loc = Convert_env.get_loc env in
  let thisNode = mk_expr env (VARIABLE "this") in
  let create_stmt_set_shift_vmt (l, previous_index)
      { base = base_class_name;
        templated_kind = t;
        is_virtual = _ (* is_base_virtual *);
        vmt_position = index;
      }
    =
    if (index = previous_index)
    then (l, previous_index)
    else begin
      (* *((_frama_c_vmt ** ) & this->_frama_c__Z1B)
               = & __::_frama_c_vmt_header_for_shift_...; *)
      let base_class_name, t =
        Convert_env.typedef_normalize env base_class_name t
      in
      let field_name =
        "_frama_c" ^ Mangling.mangle base_class_name t None
      in
      let dst =
        mk_expr env
          (UNARY(
              MEMOF,
               mk_expr env
                 (CAST
                    (([SpecType (Tstruct ("_frama_c_vmt", None, []))],
                      PTR ([], PTR ([], JUSTBASE))),
                     SINGLE_INIT
                       (mk_expr env
                          (UNARY
                             (ADDROF,
                              mk_expr env
                                (MEMBEROFPTR
                                   (thisNode,field_name)))))))))
      in
      let vmt_header_for_shift = 
        { prequalification = qualification_class;
          decl_name =
            "_frama_c_vmt_header_for_shift_" ^ (string_of_int index) }
      in
      let vmt_header_for_shift, _ =
        Convert_env.typedef_normalize env vmt_header_for_shift TStandard
      in
      let vmt_name =
        Mangling.mangle vmt_header_for_shift TStandard None in
      let src =
        mk_expr env (UNARY(ADDROF, mk_expr env (VARIABLE vmt_name)))
      in
      { stmt_ghost = false;
        stmt_node =
          COMPUTATION(mk_expr env (BINARY(ASSIGN, dst, src)), loc)} :: l,
      index
    end
  in
  (* *((_frama_c_vmt ** )this) = & ...::_frama_c_vmt_header; *)
  let dst =
    mk_expr env (
      UNARY(
        MEMOF,
        mk_expr env
          (CAST((
               [SpecType (Tstruct ("_frama_c_vmt", None, []))],
               PTR ([], PTR ([], JUSTBASE))),
               SINGLE_INIT thisNode))))
  in
  let vmt_header = 
    { prequalification = qualification_class; decl_name = "_frama_c_vmt_header"}
  in
  let vmt_header, _ = Convert_env.typedef_normalize env vmt_header TStandard in
  let vmt_name = Mangling.mangle vmt_header TStandard None in
  let src = mk_expr env (UNARY(ADDROF, mk_expr env (VARIABLE vmt_name))) in
  List.rev ({ stmt_ghost = false;
    stmt_node = COMPUTATION(mk_expr env (BINARY(ASSIGN, dst, src)), loc)}
    :: (fst (List.fold_left create_stmt_set_shift_vmt ([], 0) inherits)))

let rec add_bare_to_qualification qualif = match qualif with
  | [] -> []
  | [(QStructOrClass name)]
    -> [QStructOrClass (name ^ "_frama_c_bare")]
  | [(QTemplateInstance (name,prms))]
    -> [(QTemplateInstance ((name ^ "_frama_c_bare"), prms))]
  | x::l -> x:: (add_bare_to_qualification l)

(* functions required by the generic implicit body builder to 
   create a given implicit member function. *)
type implicit_operation = 
  { is_copy: bool;
    (* is this a copy/move operation or default constructor or destructor *)
    get_op:
      Convert_env.env -> bool -> qualified_name * tkind -> 
      (qualified_name * signature) option;
    add_op:
      Convert_env.env -> bool -> qualified_name ->
      signature -> Convert_env.env;
    signature: bool -> (qualified_name * tkind) -> signature;
    class_field: expression -> expression -> expression list;
    (* class_field dst src returns the arguments given to the
       corresponding implicit operator (in particular, default
       constructor and destructor won't use the second argument). *)
    plain_field: expression -> expression -> statement;
    (* what to do with scalar fields. assignment or nothing *)
    kind: bool -> funkind;
    name: bool -> (qualified_name * tkind) -> qualified_name;
    return: unit -> statement
      (* what to return -- nothing for constructors, this for operators. *)
  }

let remove_inherited_fields env fields (s,t) =
  let virtual_bases = Class.get_virtual_base_classes (s,t) in
  let bases = Class.get_bases_list (s,t) in
  let non_inherited (name, _) =
    List.for_all
      (fun (qn,tk) -> create_base_field_name env qn tk <> name)
      virtual_bases
      &&
      List.for_all
        (fun i ->
           create_base_field_name env i.base i.templated_kind <> name)
        bases
  in
  List.filter non_inherited fields

let rec implicit_op_call op env most_derived (s, t) dst src =
  let loc = Convert_env.get_loc env in
  let most_derived =
    most_derived || not (Class.has_virtual_base_class (s,t))
  in
  let name, signature, defs, env =
    let class_name = Convert_env.typedef_normalize env s t in
    let opname = op.name most_derived class_name in
    (* Due to the way we translate base operators for classes with virtual
       bases, op_class_name may differ from class_name for those operators.
       Failing to get them would in turn result in duplicate definitions of
       this operators if they are used more than once (i.e. in their own class
       and in another class that inherits from it).
    *)
    let op_class_name =
      Convert_env.class_name_from_qualifications env opname.prequalification
    in
    match op.get_op env most_derived op_class_name with
    | None -> create_generic_op op env most_derived class_name
    | Some (name, signature) -> name, signature, [], env
  in
  let cname =
    Mangling.mangle name TStandard (Some (op.kind most_derived, signature))
  in
  let args =
    op.class_field
      (mk_expr env (UNARY (ADDROF, dst))) (mk_expr env (UNARY (ADDROF, src)))
  in
  { stmt_ghost = Convert_env.is_ghost env;
    stmt_node =
      COMPUTATION(
        mk_expr env
          (CALL(mk_expr env (VARIABLE cname), args)), loc) }, defs, env

and create_generic_op op env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let is_union =
    match Convert_env.get_aggregate env class_name with
    | (CClass | CStruct),_ -> false
    | CUnion, _ -> true
  in
  let is_destructor =
    match op.kind most_derived with
    | FKDestructor _ -> true
    | _ -> false
  in
  let mytype = Convert_env.struct_or_union env class_name in
  let has_virtual = Class.has_virtual_base_class class_name in
  let signature = op.signature most_derived class_name in
  let signature = Convert_env.signature_normalize env signature in
  let name = op.name most_derived class_name in
  let env = op.add_op env most_derived name signature in
  let cname =
    Mangling.mangle name TStandard (Some (op.kind most_derived, signature))
  in
  let this_arg =
    { arg_type = Cxx_utils.(obj_ptr (unqual_type mytype));
      arg_name = "this"; arg_loc = loc }
  in
  let arg_name = "__frama_c_arg_0" in
  let args =
    match signature.parameter with
    | [] -> (* default constructor or destructor. *) [ this_arg ]
    | [ arg_type ] -> (* copy/move constructor/assign operator *)
      [ this_arg;
        { arg_name; arg_type; arg_loc = loc } ]
    | _ ->
      Convert_env.fatal
        env "Unexpected number of arguments for an implicit member function"
  in
  let (rt, defname) =
    make_prototype
      (Convert_env.get_loc env) env cname signature.result
      args signature.variadic false
  in
  let ret_stmt = op.return () in
  let body, defs, env =
    if is_union then begin
      (* for union, this is simply a memcopy of the whole object, see 12.8§17.*)
      if op.is_copy then begin
        let dst = mk_expr env (UNARY(MEMOF, mk_expr env (VARIABLE "this"))) in
        let src = mk_expr env (UNARY(MEMOF, mk_expr env (VARIABLE arg_name))) in
        let stmt, defs, env =
          create_assign_stmt_type op env (Cxx_utils.unqual_type mytype) dst src
        in
        raw_block [ stmt; ret_stmt ], defs, env
      end else begin
        (* non-scalar members have only trivial operations, otherwise
           the corresponding operation of the union would be deleted, as
           per 9.5§1. Hence, the operation is basically a no-op.
        *)
        raw_block [ ret_stmt ], [], env
      end
    end else if has_virtual && most_derived then begin
      let virtual_fields = Class.get_virtual_base_classes class_name in
      (* most derived operator for a class with virtual base(s): first
         call operators of virtual bases, then call operator of bare class.*)
      let virtual_base_op (stmts, defs, env) (qname, tkind) =
        let name, tkind = Convert_env.typedef_normalize env qname tkind in
        let field_name = create_base_field_name env name tkind in
        let dst =
          mk_expr env (MEMBEROFPTR(mk_expr env (VARIABLE "this"), field_name))
        in
        let src =
          mk_expr
            env (MEMBEROFPTR (mk_expr env (VARIABLE arg_name), field_name))
        in
        let stmt, defs', env =
          implicit_op_call op env true (name, tkind) dst src
        in
        stmt::stmts, defs @ defs', env
      in
      (* We first initialize the virtual bases, then the rest of the class
         as usual. See 12.6.2§10 *)
      let stmts, defs, env =
        List.fold_left virtual_base_op ([],[],env) virtual_fields
      in
      let own_stmt, defs', env =
        implicit_op_call
          op env false class_name
          (mk_expr env (UNARY(MEMOF, (mk_expr env (VARIABLE "this")))))
          (mk_expr env (UNARY(MEMOF, (mk_expr env (VARIABLE arg_name)))))
      in
      let stmts =
        (* For destructor, virtual bases are destroyed after
           the complete class. See 12.4§7. *)
        if is_destructor then own_stmt :: stmts @ [ret_stmt ]
        else List.rev_append stmts [ own_stmt; ret_stmt ]
      in
      raw_block stmts, defs @ defs', env
    end else begin
      let inherits = Class.get_bases_list class_name in
      let own_fields = Convert_env.get_struct env class_name in
      let own_fields = remove_inherited_fields env own_fields class_name in
      let dst fname =
        mk_expr env (MEMBEROFPTR (mk_expr env (VARIABLE "this"), fname))
      in
      let src fname =
        mk_expr env (MEMBEROFPTR (mk_expr env (VARIABLE arg_name), fname))
      in
      (* no virtual classes or not most derived operator: just do
           the inherited ops, vmt for current class, then own fields ops *)
      let base_op (stmts, defs, env as acc) inh =
        let field_name =
          create_base_field_name env inh.base inh.templated_kind
        in
        let src = src field_name in
        let dst = dst field_name in
        match inh.is_virtual with
        | VVirtual -> acc
        | VStandard ->
          let stmt, defs', env =
            implicit_op_call
              op env false (inh.base, inh.templated_kind) dst src
          in
          stmt :: stmts, defs @ defs', env
      in
      let stmts, defs, env =
        List.fold_left base_op ([], [], env) inherits
      in
      let stmts =
        (* For default constructor and destructor: position vmt accordingly. *)
        if not op.is_copy &&
           Convert_env.struct_has_virtual env class_name
        then begin
          let quals =
            Convert_env.get_namespace
              (Convert_env.set_namespace_from_class env class_name)
          in
          create_stmt_set_vmt env quals inherits @ stmts
        end else stmts
      in
      let stmts = if is_destructor then List.rev stmts else stmts in
      let field_op (stmts, defs, env) (fname, ftype) =
        let src = src fname in
        let dst = dst fname in
        let stmt, defs', env = create_assign_stmt_type op env ftype dst src in
        stmt::stmts, defs @ defs', env
      in
      let stmts, defs, env =
        List.fold_left field_op (stmts, defs, env) own_fields
      in
      raw_block (List.rev_append stmts [ret_stmt]), defs, env
    end
  in
  let def =
    Convert_env.is_ghost env,
    FUNDEF(None, (rt, defname), body,loc,loc)
  in
  name, signature, def::defs, env

and create_assign_stmt_type op env typ dst src =
  let loc = Convert_env.get_loc env in
  match typ.plain_type with
    | Void | Int _ | Enum _ | Float _
    | Pointer _ | LVReference _ | RVReference _ ->
      op.plain_field dst src, [], env
    | Array { subtype; dimension = Some dim } ->
        let mk_subobj env idx_var =
          let src = mk_expr env (INDEX (src,idx_var)) in
          let dst = mk_expr env (INDEX (dst,idx_var)) in
          create_assign_stmt_type op env subtype dst src
        in
        let _,_,cdim = convert_expr env [] dim false in
        let idx_name = Convert_env.temp_name env "idx" in
        let lenv = Convert_env.add_local_var env idx_name (Int IULong) in
        let stmt, defs, lenv =
          iter_on_array lenv idx_name cdim mk_subobj
        in
        stmt, defs, Convert_env.unscope lenv env
    | Array { dimension = None} ->
        (* GNU extension. Treat it as a an empty field. *)
        { stmt_ghost = false; stmt_node = NOP loc }, [], env
    | Union (s,t) ->
      (* TODO: check whether the union has a user-provided copy constructor.
         Otherwise, the implicit default is to copy the object representation
         (C++11, [class.copy]§17)
      *)
      if op.is_copy then begin
        let s, t = Convert_env.typedef_normalize env s t in
        let ctyp = Tunion (Mangling.mangle s t None,None,[]) in
        { stmt_ghost = Convert_env.is_ghost env;
          stmt_node =
            COMPUTATION
              (mk_expr env
                 (CALL(
                     mk_expr env (VARIABLE "Frama_C_memcpy"),
                     [mk_expr env (UNARY(ADDROF,dst));
                      mk_expr env (UNARY(ADDROF,src));
                      mk_expr env (TYPE_SIZEOF([SpecType ctyp],JUSTBASE))])),
               loc)}, [], env
      end else
        (* No initialization is performed by default constructor, cf 12.6.2§8.
           No destructor call for variant members, cf 12.4§7.
        *)
        { stmt_ghost = Convert_env.is_ghost env; stmt_node = NOP loc }, [], env
    | Struct (s,t) ->
      let class_name = Convert_env.typedef_normalize env s t in
      implicit_op_call op env true class_name dst src
    | Named (s,_) ->
      let typ = (Convert_env.get_typedef env s).plain_type in
      let qtyp = { qualifier = []; plain_type = typ } in
      create_assign_stmt_type op env qtyp dst src

let op_name opname most_derived (qname,tkind) =
  let qname = Cxx_utils.meth_name qname tkind opname in
  if most_derived then qname
  else
    { qname with
      prequalification = add_bare_to_qualification qname.prequalification }
       
let create_default_constructor env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let get_op env most_derived =
    if most_derived then
      Convert_env.get_option_default_constructor env
    else
      Convert_env.get_option_default_constructor_base env
  in
  let add_op env most_derived =
    if most_derived then
      Convert_env.add_default_constructor env
    else
      Convert_env.add_default_constructor env
  in
  let class_field dst _ = [ dst ] in
  (* TODO: add in intermediate_format.ast information about initializer
     of corresponding fields. See 12.6.2§8 *)
  let plain_field _ _ =
    { stmt_ghost = Convert_env.is_ghost env; stmt_node = NOP loc }
  in
  let kind most_derived = FKConstructor most_derived in
  let name _ (qname, tkind) =
    Cxx_utils.meth_name qname tkind qname.decl_name
  in
  let signature _ _ =
    { result = Cxx_utils.unqual_type Void; parameter = []; variadic = false }
  in
  let return () =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = RETURN ({ expr_loc = loc; expr_node = NOTHING}, loc) }
  in
  create_generic_op
    { is_copy = false;
      get_op; add_op; class_field; plain_field; kind; name; signature; return }
    env most_derived class_name

let create_copy_constructor env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let get_op env most_derived =
    if most_derived then
      Convert_env.get_option_copy_constructor env
    else
      Convert_env.get_option_copy_constructor_base env
  in
  let add_op env most_derived =
    if most_derived then
      Convert_env.add_copy_constructor env
    else
      Convert_env.add_copy_constructor_base env
  in
  let kind most_derived = FKConstructor most_derived in
  let name _ (qname, tkind) =
    Cxx_utils.meth_name qname tkind qname.decl_name
  in
  let class_field dst src = [ dst; src ] in
  let plain_field dst src =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = COMPUTATION(mk_expr env (BINARY(ASSIGN,dst,src)), loc)}
  in
  let signature _ class_name =
    let result = Cxx_utils.unqual_type Void in
    let mytype = Convert_env.struct_or_union env class_name in
    let parameter = [ Cxx_utils.(obj_lvref (const_type mytype)) ] in
    { result; parameter; variadic = false }
  in
  let return () =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = RETURN ({ expr_loc = loc; expr_node = NOTHING}, loc) }
  in
  create_generic_op
    { is_copy = true;
      get_op; add_op; kind; name; signature; class_field; plain_field; return }
    env most_derived class_name

let create_move_constructor env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let get_op env most_derived =
    if most_derived then
      Convert_env.get_option_move_constructor env
    else
      Convert_env.get_option_move_constructor_base env
  in
  let add_op env most_derived =
    if most_derived then
      Convert_env.add_move_constructor env
    else
      Convert_env.add_move_constructor_base env
  in
  let kind most_derived = FKConstructor most_derived in
  let name _ (qname, tkind) =
    Cxx_utils.meth_name qname tkind qname.decl_name
  in
  let class_field dst src = [ dst; src ] in
  let plain_field dst src =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = COMPUTATION(mk_expr env (BINARY(ASSIGN,dst,src)), loc)}
  in
  let signature _ class_name =
    let result = Cxx_utils.unqual_type Void in
    let mytype = Convert_env.struct_or_union env class_name in
    let parameter = [Cxx_utils.(obj_rvref (unqual_type mytype))] in
    { result; parameter; variadic = false }
  in
  let return () =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = RETURN ({ expr_loc = loc; expr_node = NOTHING}, loc) }
  in
  create_generic_op
    { is_copy = true;
      get_op; add_op; kind; name; signature; class_field; plain_field; return }
    env most_derived class_name

let create_assign_operator env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let get_op env most_derived =
    if most_derived then
      Convert_env.get_option_assign_operator env
    else
      Convert_env.get_option_assign_operator_base env
  in
  let add_op env most_derived =
    if most_derived then
      Convert_env.add_assign_operator env
    else
      Convert_env.add_assign_operator_base env
  in
  let kind _ = FKMethod [] in
  let name = op_name "operator=" in
  let class_field dst src = [ dst; src ] in
  let plain_field dst src =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = COMPUTATION(mk_expr env (BINARY(ASSIGN,dst,src)), loc)}
  in
  let signature _ class_name =
    let mytype = Convert_env.struct_or_union env class_name in
    let result = Cxx_utils.(obj_lvref (unqual_type mytype)) in
    let parameter = [Cxx_utils.(obj_lvref (const_type mytype))] in
    { result; parameter; variadic = false }
  in
  let return () =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node =
        RETURN ({expr_loc = loc; expr_node = VARIABLE "this"}, loc) }
  in
  create_generic_op
    { is_copy = true;
      get_op; add_op; kind; name; signature; class_field; plain_field; return }
    env most_derived class_name

let create_move_operator env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let get_op env most_derived =
    if most_derived then
      Convert_env.get_option_move_operator env
    else
      Convert_env.get_option_move_operator_base env
  in
  let add_op env most_derived =
    if most_derived then
      Convert_env.add_move_operator env
    else
      Convert_env.add_move_operator_base env
  in
  let kind _ = FKMethod [] in
  let name = op_name "operator=" in
  let class_field dst src = [ dst; src ] in
  let plain_field dst src =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = COMPUTATION(mk_expr env (BINARY(ASSIGN,dst,src)), loc)}
  in
  let signature _ class_name =
    let mytype = Convert_env.struct_or_union env class_name in
    let result = Cxx_utils.(obj_lvref (unqual_type mytype)) in
    let parameter = [Cxx_utils.(obj_rvref (unqual_type mytype))] in
    { result; parameter; variadic = false }
  in
  let return () =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = RETURN ({ expr_loc = loc; expr_node = VARIABLE "this"}, loc)}
  in
  create_generic_op
    { is_copy = true;
      get_op; add_op; kind; name; signature; class_field; plain_field; return }
    env most_derived class_name

let create_destructor env most_derived class_name =
  let loc = Convert_env.get_loc env in
  let add_op env most_derived class_name _ =
    if most_derived then
      Convert_env.add_destructor env class_name.prequalification
    else
      Convert_env.add_destructor_base env class_name.prequalification
  in
  let name _ (s,t) = Cxx_utils.meth_name s t ("~" ^ s.decl_name) in
  let signature _ _ =
    { result = Cxx_utils.unqual_type Void; parameter = []; variadic = false }
  in
  (* TODO: Handle the case where the destructor is virtual *)
  let get_op env most_derived class_name =
    let has_dest =
      if most_derived then
        Convert_env.has_destructor env class_name
      else
        Convert_env.has_destructor_base env class_name
    in
    if has_dest then
      Some (name most_derived class_name, signature most_derived class_name)
    else None
  in
  let class_field dst _ = [ dst ] in
  let plain_field _ _ =
    { stmt_ghost = Convert_env.is_ghost env; stmt_node = NOP loc }
  in
  let kind most_derived = FKDestructor most_derived in
  let return () =
    { stmt_ghost = Convert_env.is_ghost env;
      stmt_node = RETURN ({ expr_loc = loc; expr_node = NOTHING}, loc) }
  in
  create_generic_op
    { is_copy = false;
      add_op; name; signature; get_op; class_field; plain_field; kind; return }
    env most_derived class_name

let create_implicit_body env kind args class_name =
  (* NB: the this parameter is present in the args list. *)
  match kind, args with
  | FKConstructor derived, [ _ ] ->
    create_default_constructor env derived class_name
  | FKConstructor d, [ _; { arg_type = { plain_type = LVReference _ } } ] ->
    create_copy_constructor env d class_name
  | FKConstructor d, [ _; { arg_type = { plain_type = RVReference _ } } ] ->
    create_move_constructor env d class_name
  | FKMethod _, [ _; { arg_type = { plain_type = (Struct _ | Union _) } }] ->
    create_assign_operator env true class_name
  | FKMethod _, [ _; { arg_type = { plain_type = LVReference _ } }] ->
      create_assign_operator env true class_name
  | FKMethod _, [ _; { arg_type = { plain_type = RVReference _ } }] ->
    create_move_operator env true class_name
  | FKDestructor d, [ _ ] ->
    create_destructor env d class_name
  | _ ->
      Frama_Clang_option.not_yet_implemented
        "Don't know this kind of implicitly defined member function"

let has_default_value _arg = false
(* TODO: export this information from clang *)

let is_assign_operator s = s = "operator="

let add_special_member env name kind rt args =
  let (class_name, tc) =
    Convert_env.class_name_from_qualifications env name.prequalification
  in
  let signature args =
    { result = rt;
      parameter = List.map (fun a -> a.arg_type) args;
      variadic = false
    }
  in
  let rec is_own_class t =
    match t.plain_type with
      | Struct (s,ts)
      | Union (s,ts)
        -> Fclang_datatype.Qualified_name.equal (s,ts) (class_name, tc)
      | Named (s,_) when Cxx_utils.is_builtin_qual_type s -> false
      | Named (s,_) -> is_own_class { qualifier = [];
                       plain_type = (Convert_env.get_typedef env s).plain_type}
      | _ -> false
  in
  match kind, args with
    | FKConstructor d, _::l when List.for_all has_default_value l ->
      (if d then Convert_env.add_default_constructor
       else Convert_env.add_default_constructor_base)
        env name (signature l)
    | FKConstructor d, 
      _::({arg_type= { plain_type = LVReference (PDataPointer t) }}::l as args)
        when is_own_class t && List.for_all has_default_value l ->
      (if d then Convert_env.add_copy_constructor
       else Convert_env.add_copy_constructor_base)
        env name (signature args)
    | FKConstructor d,
      _::({arg_type= { plain_type = RVReference (PDataPointer t) }}::l as args)
        when is_own_class t && List.for_all has_default_value l ->
      (if d then Convert_env.add_move_constructor
       else Convert_env.add_move_constructor_base)
        env name (signature args)
    | FKMethod [],_::([{arg_type = t}] as args) when
        is_own_class t && is_assign_operator name.decl_name ->
      Convert_env.add_assign_operator env name (signature args)
    | FKMethod [],
          _::([{arg_type ={plain_type = LVReference (PDataPointer t)}}] as args)
            when is_own_class t && is_assign_operator name.decl_name ->
      Convert_env.add_assign_operator env name (signature args)
    | FKMethod [],
          _::([{arg_type ={plain_type = RVReference (PDataPointer t)}}] as args)
            when is_own_class t && is_assign_operator name.decl_name ->
      Convert_env.add_move_operator env name (signature args)
    | FKDestructor d, _ ->
      (if d then Convert_env.add_destructor
       else Convert_env.add_destructor_base)
        env name.prequalification
    | _ -> env

let add_arg_names l =
  let new_name idx = if idx = -1 then "x" else "x" ^ (string_of_int idx) in
  List.rev 
    (fst
       (List.fold_left
          (fun (acc,idx) arg ->
            if arg.arg_name = "" then
              { arg with arg_name = new_name idx }::acc, idx+1
            else arg::acc,idx)
          ([],(-1))
          l))

let convert_contract env spec =
  Convert_acsl.convert_function_contract env spec,
  Convert_env.get_loc env

let rec collect_class_components env body =
  List.fold_left
    convert_class_component (env,[],[],[],[]) (reorder_implicit body)

(* Note: the list of others global definition must be built in reverse
   order, as it will be reverted by the callers. *)
and convert_class_component (env, implicits, types, fields, others) meth =
  let ghost = Convert_env.is_ghost env in
  match meth with
    | CMethod(loc,name,kind,return_type,all_args,variadic,body,implicit,tkind,
              has_further_definition,_(*throws*),spec) ->
      let class_name =
        Convert_env.(class_name_from_qualifications env (get_namespace env))
      in
      let qname = Convert_env.qualify env name in
      let env = Convert_env.set_loc env loc in
      let env = add_special_member env qname kind return_type all_args in
      let all_args = if implicit then add_arg_names all_args else all_args in
      let args = remove_this_parameter kind all_args in
      let args_sig = List.map (fun x -> x.arg_type) args in
      let benv = Convert_env.add_formal_parameters env all_args in
      let benv = Convert_env.set_current_func_name benv qname in
      let benv =
        Convert_env.set_current_return_type benv return_type.plain_type
      in
      let qname,tkind= Convert_env.typedef_normalize env qname tkind in
      let signature = {result= return_type; parameter= args_sig; variadic } in
      let signature = Convert_env.signature_normalize env signature in
      let name = Mangling.mangle qname tkind (Some(kind,signature)) in
      let spec = Extlib.opt_map (convert_contract benv) spec in
      let extern_c = false in
      let spec =
        if has_further_definition
        then spec
        else
          Generate_spec.add_contract ~env:benv ~kind ~return_type ~args
            ~variadic ~implicit ~extern_c spec
      in
      let (rt,name as defname) =
        make_prototype loc benv name return_type all_args variadic false
      in
      let has_virtual_base = Class.has_virtual_base_class class_name in
      (match body with
        | None ->
          let implicits =
            if implicit && not has_further_definition
            then (kind,all_args)::implicits
            else implicits
          in
          (env,implicits, types, fields,
           (ghost,DECDEF(spec,(rt,[name,NO_INIT]),loc))::others)
        | Some body ->
          if (not has_virtual_base)
            || Frama_Clang_option.Cxx_virtual_bare_methods_in_clang.get ()
          then 
            let body = convert_block benv body false in
              (env,implicits,types,fields,
               (ghost, FUNDEF(spec,defname,body,loc,loc))
              ::others)
          else
            let cbody = convert_block benv body false in
            let cbody_bare = convert_block benv body true in
            let defname_bare =
              make_prototype loc benv
                (Mangling.mangle 
                  { qname with prequalification
                    = add_bare_to_qualification qname.prequalification } 
                  tkind (Some(kind,signature)))
                return_type all_args variadic false
            in
              (env,implicits,types,fields,
               (ghost, FUNDEF(spec,defname,cbody,loc,loc))
               ::(ghost, FUNDEF(spec,defname_bare,cbody_bare,loc,loc))::others))
    | CCompound(loc,name,kind,inherits,body,tkind,has_virtual) ->
      let new_env = Convert_env.set_loc env loc in
      let qualified_name = Convert_env.qualify new_env name in
      let new_env = make_class_env new_env (Declaration name) kind tkind false
      in let new_env,new_implicits,subtypes,subfields,subothers =
        match body with
          | None -> new_env, [], [], [], []
          | Some body -> collect_class_components new_env body
      in
      let new_env, class_decl, bare_class_decl =
        make_class_decl
          new_env qualified_name tkind kind inherits subfields body has_virtual
      in
      let qualified_name, tkind =
        Convert_env.typedef_normalize env qualified_name tkind
      in
      let new_env, my_implicits =
        List.fold_left
          (fun (env, acc) (kind, args) ->
             let _, _, defs, env =
               create_implicit_body env kind args (qualified_name, tkind)
             in
             env, acc @ defs)
          (new_env, []) new_implicits
      in
      let subothers = my_implicits @ subothers in
      Class.add_class (qualified_name,tkind);
      Extlib.may
        (List.iter (Class.add_inheritance_relation(qualified_name, tkind)))
        inherits;
      Convert_env.copy_namespace new_env env,
      implicits,
      (match bare_class_decl with
        | None -> class_decl::(List.append subtypes types)
        | Some bare_class_decl -> bare_class_decl::(class_decl
            ::(List.append subtypes types))),
      fields,
      List.append subothers others
    | CFieldDecl(loc, name, typ, bf_info) ->
      let env = Convert_env.set_loc env loc in
      let (base,decl) = convert_specifiers env typ false in
      (* We remove static qualifier to avoid confusion with the C's static.
         We remove const qualifier to allow the Ctor to write the attribute.
         There is no problem of semantic for call of overloaded functions
         because the overloading is resolved by FramaCIRGen and consequently,
         the correct function to call is specified in the Fir.*)
      let base =
        List.filter
          (function
            | SpecStorage(STATIC) | SpecCV(CV_CONST) -> false | _ -> true)
          base
      in
      if is_static typ.qualifier then begin
        let attrs = add_fc_destructor_attr env typ [] in
        let name = Convert_env.qualify env name in
        let cname,_= Convert_env.typedef_normalize env name TStandard in
        let cname = Mangling.mangle cname TStandard None in
        Convert_env.add_global_var env name typ.plain_type,
        implicits,
        types,
        fields,
        (ghost,
         DECDEF(
             None,
             (base, [(cname, decl JUSTBASE, attrs,loc),NO_INIT]), loc))
        ::others
      end else begin
        (* no need to mangle names: they can be shared by various structures *)
        let bf_length = Extlib.opt_map (convert_bitfield_info env) bf_info in
        (env,implicits,
         types,
         (FIELD(base,[(name,decl JUSTBASE,[],loc),bf_length]), typ) ::fields,
         others)
      end
    | CTypedef (loc,name,typ) ->
      let env = Convert_env.set_loc env loc in
      let qual_name = Convert_env.qualify env name in
      let env = Convert_env.add_typedef env qual_name typ in
      let name,_= Convert_env.typedef_normalize env qual_name TStandard in
      let name = Mangling.mangle name TStandard None in
      let (base,decl) = convert_specifiers env typ false in
      (env,implicits,
       (ghost,
        TYPEDEF((SpecTypedef::base,[(name,decl JUSTBASE,[],loc)]),loc))::types,
       fields,
       others)
    | CStaticConst(loc,name, ikind,_,value) ->
      (* static in the case of a class does not mean internal storage,
         but static data member. The resulting variable should not be tagged
         with static storage qualifier in C. *)
      let decl, env =
        convert_static_const env loc name ikind ICLiteral value false
      in
      (env,implicits,types,fields,decl :: others)
    | CEnum(loc,name,ikind, body) ->
      let env = Convert_env.set_loc env loc in
      let name, _ = Convert_env.typedef_normalize env
            (Convert_env.qualify env name) TStandard in
      let name = Mangling.mangle name TStandard None
      in
      let tags = Extlib.opt_map (convert_enum_constants env ikind) body in
           (env,implicits,
            (ghost, DECDEF(None,([SpecType (Tenum(name,tags,[]))],[]),loc))
            :: types, fields, others)
    | Class_annot(loc,annot) ->
      let env = Convert_env.set_loc env loc in
      let annot = Convert_acsl.convert_annot env annot in
      (env,implicits,types,fields,(ghost, GLOBANNOT [annot]):: others)

and convert_class env name tkind kind inherits body has_virtual =
  let qualified_name = make_qualified_name env name in
  let this_env = make_class_env env name kind tkind false in
  let new_env, implicits, types, fields, others =
    match body with
    | None -> this_env, [], [], [], []
    | Some body -> collect_class_components this_env body
  in
  let new_env, decl, bare_decl =
    make_class_decl
      new_env qualified_name tkind kind inherits fields body has_virtual
  in
  let new_env, my_implicits =
    List.fold_left
      (fun (env, acc) (kind, args) ->
         let _,_,defs,env =
           create_implicit_body env kind args (qualified_name, tkind)
         in
         env, List.rev_append defs acc)
      (new_env, [])
      implicits
  in
  (* implicit definitions might depend on internal defs, in particular in
     presence of nested classes that occur as data member of the current class.
  *)
  let others = List.rev_append others my_implicits in
   List.rev_append types (match bare_decl with
    | None -> decl :: others
    | Some bdecl -> bdecl :: (decl :: others)),
  Convert_env.copy_namespace new_env env

let rec convert_pod_field (cfields, typs, decls, env) field =
  match field with
  | CFieldDecl(loc, name, typ, bf_info) ->
    let env = Convert_env.set_loc env loc in
    let (base,decl) = convert_specifiers env typ false in
    let bf_length = Extlib.opt_map (convert_bitfield_info env) bf_info in
    FIELD(base, [(name,decl JUSTBASE,[],loc),bf_length])::cfields,
    (name, typ) :: typs, decls, env
  | CCompound(_, name, kind, _, body, tn, _) ->
      let dname = Declaration name in
      let cur_decls, env = convert_pod env dname kind tn body in
      cfields, typs, decls @ cur_decls, env
  | _ ->
    Convert_env.fatal env "Unknown declaration in extern C structure definition"

and convert_pod env name kind tn body =
  let loc = Convert_env.get_loc env in
  let ghost = Convert_env.is_ghost env in
  let name =
    match name with
    | Declaration name -> name
    | Implementation name -> name.decl_name
  in
  let env = Convert_env.add_aggregate env (empty_qual name) kind tn true in
  let fields, inner_decls, env =
    match body with
      | None -> None, [], env
      | Some body ->
        let fields, typs, decls, env =
          List.fold_left convert_pod_field ([],[], [], env) body
          (* List.fold_right convert_pod_field body ([],[], env) *)
        in
        Some (List.rev fields), decls,
        Convert_env.add_struct env ((empty_qual name),tn) (List.rev typs)
  in
  let type_decl =
    match kind with
      | CClass | CStruct -> Tstruct (name, fields, [])
      | CUnion -> Tunion(name, fields,[])
  in
  inner_decls @ [ghost, DECDEF(None,([SpecType type_decl],[]), loc)], env

let convert_class env name tkind kind inherits body has_virtual =
  if Convert_env.is_extern_c env then convert_pod env name kind tkind body
  else convert_class env name tkind kind inherits body has_virtual

let make_global_cons_init env name init =
  let loc = Convert_env.get_loc env in
  FUNDEF(None,([SpecType Tvoid; SpecAttr("__constructor__",[])],
               ("__fc_init" ^ name, PROTO(JUSTBASE,[],false),[],loc)),
         { blabels = []; battrs = []; bstmts = init }, loc, loc)

let builtins = [ "__builtin_va_start" ]

let do_not_translate name =
  let id =
    match name with
      | Declaration name -> name
      | Implementation name -> name.decl_name
  in
  List.mem id builtins

let extract_name = function
  | FUNDEF(_,(_,(name,_,_,_)),_,_,_) -> name
  | DECDEF(_,(_,((name,_,_,_),_)::_),_) -> name
  | DECDEF _ -> Frama_Clang_option.fatal "Empty declaration"
  | TYPEDEF((_,((name,_,_,_)::_)),_) -> name
  | TYPEDEF _ -> Frama_Clang_option.fatal "Empty declaration"
  | ONLYTYPEDEF _ -> Frama_Clang_option.fatal "Empty declaration"
  | GLOBASM _ | PRAGMA _ | LINKAGE _ | GLOBANNOT _ | CUSTOM _ ->
    Frama_Clang_option.fatal "Empty declaration"

let add_glob_temp =
  let count = ref 0 in
  let anonymous_name () =
    let res = if !count = 0 then "anon" else "anon" ^ string_of_int !count in
    incr count; res
  in
  fun env defs (def,init_stmt) ->
    let ghost = Convert_env.is_ghost env in 
    let defs =
      match init_stmt with
        | None -> defs
        | Some stmt ->
          let name =
            match def with
              | None -> anonymous_name ()
              | Some d -> extract_name d
          in
          let init = make_global_cons_init env name [stmt] in
          (ghost,init)::defs
  in
  match def with None -> defs | Some d -> (ghost,d)::defs

let is_frama_clang_array_init_name = function
  | Implementation { decl_name } ->
    let prefix = "__fc_init_array" in
    let l = String.length prefix in 
    l <= String.length decl_name && String.sub decl_name 0 l = prefix
  | Declaration _ -> false 
(* NB: we should have a loc for the declaration itself
   as well as for the body*)
let rec convert_global env glob =
  match glob with
    | GlobalAnnotation(loc,annot) ->
      let env = Convert_env.set_loc env loc in
      let annot = Convert_acsl.convert_annot env annot in
      [false, GLOBANNOT [annot]], env
    | Function (name, _, _, _, _, _, _, _, _, _, _, _, _)
        when do_not_translate name
      -> [], env
    | Function(name, kind, loc, return_type,all_args,
               body, extern_c, ghost, variadic, tkind, has_further_definition,
               _ (* throws *), spec) ->
      let attrs = if is_frama_clang_array_init_name name then
          [ SpecAttr("__constructor__", [])]
        else []
      in
      let env = Convert_env.set_loc env loc in
      let benv = Convert_env.set_extern_c env extern_c in
      let benv = Convert_env.set_ghost benv ghost in
      let args = remove_this_parameter kind all_args in
      let args_sig = List.map (fun x -> x.arg_type) args in
      let benv = Convert_env.add_formal_parameters benv all_args in
      let qualified_name = make_qualified_name env name in
      let benv = Convert_env.set_current_func_name benv qualified_name in
      let benv =
        Convert_env.set_current_return_type benv return_type.plain_type
      in
      let qname =
        if extern_c then qualified_name.decl_name
        else
          let qualified_name, tkind = Convert_env.typedef_normalize env
                qualified_name tkind in
          let signature
            = { result = return_type; parameter = args_sig; variadic } in
          let signature = Convert_env.signature_normalize env signature in
          Mangling.mangle qualified_name tkind (Some (kind, signature))
      in
      let spec = Extlib.opt_map (convert_contract benv) spec in
      let implicit = false in
      let spec =
        if has_further_definition then spec
        else
          Generate_spec.add_contract
            ~env:benv ~kind ~return_type ~args ~variadic ~implicit ~extern_c spec
      in
      let (rt, name) =
        make_prototype loc benv qname return_type all_args variadic false
      in
      let get_class_from_this_type t =
        match t.plain_type with
          | Pointer (PDataPointer { plain_type = Struct (s,ts) | Union(s,ts) }) -> (s,ts)
          | _ -> Frama_Clang_option.fatal "bad type for this"
      in let has_virtual_base =
        match kind with
          | FKConstructor _ -> Class.has_virtual_base_class
              (get_class_from_this_type (List.hd all_args).arg_type)
          | FKDestructor _ -> Class.has_virtual_base_class
              (get_class_from_this_type (List.hd all_args).arg_type)
          | _ -> false
      in
      let rt = rt @ attrs in
      (match body with
        | None ->
          [ghost, DECDEF(spec,(rt,[name,NO_INIT]), loc)], env
        | Some body ->
          if (not has_virtual_base)
            || Frama_Clang_option.Cxx_virtual_bare_methods_in_clang.get ()
          then 
            let body = convert_block benv body false in
            [ghost, FUNDEF(spec,(rt,name),body,loc,loc)], env
          else
            let cbody = convert_block benv body false in
            let cbody_bare = convert_block benv body true in
            let qname_bare =
              let qualified_name, tkind = Convert_env.typedef_normalize env
                    qualified_name tkind in
              let signature
                = { result = return_type; parameter = args_sig; variadic } in
              let signature = Convert_env.signature_normalize env signature in
              Mangling.mangle
                { qualified_name with prequalification
                  = add_bare_to_qualification qualified_name.prequalification }
                tkind (Some (kind, signature))
            in let (rt_bare, name_bare) =
              make_prototype
                loc benv qname_bare return_type all_args variadic false
            in
            [ghost, FUNDEF(spec,(rt,name),cbody,loc,loc);
             ghost, FUNDEF(spec,(rt_bare,name_bare),cbody_bare,loc,loc)], env)
    | Compound(loc,name,kind,inherits,body,has_virtual,tc, is_extern_c,ghost) ->
      let old_ghost = Convert_env.is_ghost env in
      let env = Convert_env.set_loc env loc in
      let env = Convert_env.set_extern_c env is_extern_c in
      let env = Convert_env.set_ghost env ghost in
      let qualified_name = 
        match name with
          | Declaration n -> Convert_env.qualify env n
          | Implementation n -> n
      in let qualified_name, tc
        = Convert_env.typedef_normalize env qualified_name tc in
      Class.add_class (qualified_name,tc);
      Extlib.may
        (List.iter (Class.add_inheritance_relation(qualified_name,tc)))
        inherits;
      let globs,env=convert_class env name tc kind inherits body has_virtual in
      globs, Convert_env.set_ghost env old_ghost
    | GlobalVarDecl(loc,name,typ,init,is_extern_c, ghost) ->
      let env = Convert_env.set_loc env loc in
      let env = Convert_env.set_extern_c env is_extern_c in
      let qualified_name = make_qualified_name env name in
      let name =
        if is_extern_c
        then qualified_name.decl_name
        else
          let n', t' = Convert_env.typedef_normalize env
                qualified_name TStandard in
          Mangling.mangle n' t' None
      in
      let rt, decl = convert_specifiers env typ false in
      let attrs = add_fc_destructor_attr env typ [] in
      let env = Convert_env.add_global_var env qualified_name typ.plain_type in
      let old_ghost = Convert_env.is_ghost env in
      let env = Convert_env.set_ghost env ghost in
      let init, init_func =
        match init with
        | None -> NO_INIT, []
        | Some init ->
          let var = Global qualified_name in
          let env, tmps, init, e = convert_initializer env typ var init false in
          (match tmps, e with
            | [], None -> init, []
            | [], Some stmt ->
              NO_INIT, [ghost, make_global_cons_init env name [stmt]]
            | tmps, Some stmt ->
              assert (init = NO_INIT);
              let glob_init = make_global_cons_init env name [stmt] in
              let init_body =
                List.fold_left (add_glob_temp env) [ghost, glob_init] tmps
              in
              NO_INIT, init_body
            | tmps, None ->
              let var = { expr_loc = loc; expr_node = VARIABLE name } in
              let init_body = mk_compound_init env var typ init in
              let init_func =
                match init_body with
                | NOP _ -> []
                | BLOCK({ bstmts }, _,_) ->
                  [ghost, make_global_cons_init env name bstmts]
                | _ ->
                  [ghost, make_global_cons_init
                     env name [{stmt_ghost = ghost; stmt_node = init_body}]]
              in
              let init_all =
                List.fold_left (add_glob_temp env) init_func tmps
              in
              NO_INIT, init_all
          )
      in
      (ghost, DECDEF(None,(rt,[(name,decl JUSTBASE,attrs,loc),init]),loc))::
        init_func,
      Convert_env.set_ghost env old_ghost
    (* Just don't pollute the C file with these types. *)
    | Typedef(_,Declaration(t),_,_, _)
        when Cxx_utils.is_builtin_type t -> [], env
    | Typedef(loc,name,typ,is_extern_c_context, ghost) ->
      let env = Convert_env.set_loc env loc in
      let qual_name = make_qualified_name env name in
      let env = Convert_env.add_typedef env qual_name typ in
      let env =
        if is_extern_c_context then begin
          match name with
            | Declaration decl_name ->
                (* also add the unqualified version of the type
                   in the environment as it is also possible to use it
                   without qualification. This is in particular the case
                   for the typed imported from the C standard library
                   by the STL through the <c...> headers *)
                Convert_env.add_typedef
                  env { prequalification = []; decl_name} typ
            | _ ->
                Convert_env.fatal env
                  "extern C typedef with a fully qualified name"
        end else env
      in
      let name =
        if is_extern_c_context
        then qual_name.decl_name
        else
          let qual_name, t = Convert_env.typedef_normalize env
                qual_name TStandard in
          Mangling.mangle qual_name t None
      in
      let (base,decl) = convert_specifiers env typ false in
      [ghost,
       TYPEDEF((SpecTypedef::base,[(name,decl JUSTBASE,[],loc)]), loc)], env
    | Namespace (_,name,body) ->
      let new_env = Convert_env.add_namespace env (QNamespace name) in
      let globs, new_env =
        List.fold_left
          (fun (acc, env) g ->
            let globs, env = convert_global env g in
            globs::acc, env)
          ([], new_env)
          body
      in
      let globs = List.fold_left (fun acc l -> l @ acc) [] globs in
      globs, Convert_env.copy_namespace new_env env
    | StaticConst(loc,name,ikind,kind,value, is_extern_c, ghost) ->
        let old_ghost = Convert_env.is_ghost env in
        let env = Convert_env.set_ghost env ghost in
        let old_extern_c = Convert_env.is_extern_c env in
        let env = Convert_env.set_extern_c env is_extern_c in
        let decl, env =
          convert_static_const env loc name ikind kind value false
        in
        let env = Convert_env.set_ghost env old_ghost in
        let env = Convert_env.set_extern_c env old_extern_c in
        [decl], env
    | EnumDecl(loc,name,ikind, body,is_extern_c_context,ghost) ->
      let env = Convert_env.set_loc env loc in
      let env = Convert_env.set_extern_c env is_extern_c_context in
      let qualified_name = make_qualified_name env name in
      let name =
        if is_extern_c_context
        then qualified_name.decl_name
        else
          let qualified_name, t = Convert_env.typedef_normalize env
                qualified_name TStandard in
          Mangling.mangle qualified_name t None
      in
      let tags = Extlib.opt_map (convert_enum_constants env ikind) body in
          [ghost,
           DECDEF(None,([SpecType (Tenum(name,tags,[]))],[]),loc)], env

let convert_ast file =
  let convert_global (acc, env) g =
    let globs, env = convert_global env g in List.rev_append globs acc, env
  in
  let globs, _ =
    List.fold_left convert_global ([], Convert_env.empty_env) file.globals
  in let basic_pos =
    { Lexing.pos_fname = ""; Lexing.pos_lnum = 0;
      Lexing.pos_bol = 0; Lexing.pos_cnum = 0;
    } in
  (file.filename, (false, DECDEF (None, ([SpecType Tvoid],
    [(("malloc", PTR ([], PROTO ( JUSTBASE ,
      [((if Cil.theMachine.kindOfSizeOf == Cil_types.IULong
         then [SpecType Tunsigned; SpecType Tlong]
         else [SpecType Tunsigned]),
        ("size", JUSTBASE, [], (basic_pos, basic_pos)))],
      false)), [], (basic_pos, basic_pos)), NO_INIT)]), (basic_pos, basic_pos))) 
    :: ((false, DECDEF (None, ([SpecType Tvoid],
    [(("free", PROTO ( JUSTBASE ,
      [([SpecType Tvoid],
        ("ptr", PTR ([], JUSTBASE), [], (basic_pos, basic_pos)))],
      false), [], (basic_pos, basic_pos)), NO_INIT)]), (basic_pos, basic_pos)))
    :: ((false, DECDEF (None, ([SpecType Tvoid],
    [(("Frama_C_memcpy", PTR ([], PROTO ( JUSTBASE ,
      [([SpecType Tvoid],
        ("dest", PTR ([], JUSTBASE), [], (basic_pos, basic_pos)));
       ([SpecCV CV_CONST; SpecType Tvoid],
        ("src", PTR ([], JUSTBASE), [], (basic_pos, basic_pos)));
       (spec_of_ikind Cil.(theMachine.kindOfSizeOf),
        ("size", JUSTBASE, [], (basic_pos, basic_pos)))
      ],
      false)), [], (basic_pos, basic_pos)), NO_INIT)]), (basic_pos, basic_pos)))
    :: (List.rev globs))))
